from __future__ import annotations

import shutil

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import pytest

from packaging.utils import canonicalize_name

from poetry.core.constraints.version import parse_constraint
from poetry.core.factory import Factory
from poetry.core.packages.dependency import Dependency
from poetry.core.packages.dependency_group import MAIN_GROUP
from poetry.core.packages.directory_dependency import DirectoryDependency
from poetry.core.packages.file_dependency import FileDependency
from poetry.core.packages.url_dependency import URLDependency
from poetry.core.packages.vcs_dependency import VCSDependency
from poetry.core.pyproject.tables import BuildSystem
from poetry.core.utils._compat import tomllib
from poetry.core.version.markers import SingleMarker


if TYPE_CHECKING:
    from pytest import LogCaptureFixture


fixtures_dir = Path(__file__).parent / "fixtures"


@pytest.fixture
def complete_legacy_warnings() -> list[str]:
    return [
        "[tool.poetry.name] is deprecated. Use [project.name] instead.",
        (
            "[tool.poetry.version] is set but 'version' is not in "
            "[project.dynamic]. If it is static use [project.version]. If it "
            "is dynamic, add 'version' to [project.dynamic].\n"
            "If you want to set the version dynamically via `poetry build "
            "--local-version` or you are using a plugin, which sets the "
            "version dynamically, you should define the version in "
            "[tool.poetry] and add 'version' to [project.dynamic]."
        ),
        "[tool.poetry.description] is deprecated. Use [project.description] instead.",
        (
            "[tool.poetry.readme] is set but 'readme' is not in "
            "[project.dynamic]. If it is static use [project.readme]. If it "
            "is dynamic, add 'readme' to [project.dynamic].\n"
            "If you want to define multiple readmes, you should define them "
            "in [tool.poetry] and add 'readme' to [project.dynamic]."
        ),
        "[tool.poetry.license] is deprecated. Use [project.license] instead.",
        "[tool.poetry.authors] is deprecated. Use [project.authors] instead.",
        "[tool.poetry.maintainers] is deprecated. Use [project.maintainers] instead.",
        "[tool.poetry.keywords] is deprecated. Use [project.keywords] instead.",
        (
            "[tool.poetry.classifiers] is set but 'classifiers' is not in "
            "[project.dynamic]. If it is static use [project.classifiers]. If it "
            "is dynamic, add 'classifiers' to [project.dynamic].\n"
            "ATTENTION: Per default Poetry determines classifiers for "
            "supported Python versions and license automatically. If you "
            "define classifiers in [project], you disable the automatic "
            "enrichment. In other words, you have to define all classifiers "
            "manually. If you want to use Poetry's automatic enrichment of "
            "classifiers, you should define them in [tool.poetry] and add "
            "'classifiers' to [project.dynamic]."
        ),
        "[tool.poetry.homepage] is deprecated. Use [project.urls] instead.",
        "[tool.poetry.repository] is deprecated. Use [project.urls] instead.",
        "[tool.poetry.documentation] is deprecated. Use [project.urls] instead.",
        "[tool.poetry.plugins] is deprecated. Use [project.entry-points] instead.",
        (
            "[tool.poetry.extras] is deprecated. Use "
            "[project.optional-dependencies] instead."
        ),
        (
            "Defining console scripts in [tool.poetry.scripts] is deprecated. "
            "Use [project.scripts] instead. "
            "([tool.poetry.scripts] should only be used for scripts of type 'file')."
        ),
    ]


@pytest.fixture
def complete_legacy_duplicate_warnings() -> list[str]:
    return [
        (
            "[project.name] and [tool.poetry.name] are both set. The latter "
            "will be ignored."
        ),
        (
            "[project.version] and [tool.poetry.version] are both set. The "
            "latter will be ignored.\n"
            "If you want to set the version dynamically via `poetry build "
            "--local-version` or you are using a plugin, which sets the "
            "version dynamically, you should define the version in "
            "[tool.poetry] and add 'version' to [project.dynamic]."
        ),
        (
            "[project.description] and [tool.poetry.description] are both "
            "set. The latter will be ignored."
        ),
        (
            "[project.readme] and [tool.poetry.readme] are both set. The "
            "latter will be ignored.\n"
            "If you want to define multiple readmes, you should define them "
            "in [tool.poetry] and add 'readme' to [project.dynamic]."
        ),
        (
            "[project.license] and [tool.poetry.license] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.authors] and [tool.poetry.authors] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.maintainers] and [tool.poetry.maintainers] are both "
            "set. The latter will be ignored."
        ),
        (
            "[project.keywords] and [tool.poetry.keywords] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.classifiers] and [tool.poetry.classifiers] are both "
            "set. The latter will be ignored.\n"
            "ATTENTION: Per default Poetry determines classifiers for "
            "supported Python versions and license automatically. If you "
            "define classifiers in [project], you disable the automatic "
            "enrichment. In other words, you have to define all classifiers "
            "manually. If you want to use Poetry's automatic enrichment of "
            "classifiers, you should define them in [tool.poetry] and add "
            "'classifiers' to [project.dynamic]."
        ),
        (
            "[project.urls] and [tool.poetry.homepage] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.urls] and [tool.poetry.repository] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.urls] and [tool.poetry.documentation] are both set. "
            "The latter will be ignored."
        ),
        (
            "[project.entry-points] and [tool.poetry.plugins] are both set. The "
            "latter will be ignored."
        ),
        (
            "[project.optional-dependencies] and [tool.poetry.extras] are "
            "both set. The latter will be ignored."
        ),
        (
            "[project.scripts] is set and there are console scripts "
            "in [tool.poetry.scripts]. The latter will be ignored."
        ),
    ]


@pytest.mark.parametrize(
    "project", ["sample_project", "sample_project_new", "sample_project_dynamic"]
)
def test_create_poetry(project: str) -> None:
    new_format = project == "sample_project_new"
    dynamic = project == "sample_project_dynamic"
    poetry = Factory().create_poetry(fixtures_dir / project)

    assert poetry.is_package_mode

    package = poetry.package

    assert package.name == "my-package"
    assert package.version.text == "1.2.3"
    assert package.description == "Some description."
    assert package.authors == ["Sébastien Eustace <sebastien@eustace.io>"]
    assert package.maintainers == ["Sébastien Eustace <sebastien@eustace.io>"]
    if new_format:
        assert package.license is None
        assert package.license_expression == "MIT"
    else:
        assert package.license is not None
        assert package.license.id == "MIT"
        assert package.license_expression is None
    assert (
        package.readmes[0].relative_to(fixtures_dir).as_posix()
        == f"{project}/README.rst"
    )
    assert package.homepage == "https://python-poetry.org"
    assert package.repository_url == "https://github.com/python-poetry/poetry"
    assert package.keywords == ["packaging", "dependency", "poetry"]

    assert package.python_versions == ">=3.6"
    assert str(package.python_constraint) == ">=3.6"

    dependencies: dict[str, Dependency] = {}
    for dep in package.requires:
        dependencies[dep.name] = dep

    cleo = dependencies["cleo"]
    assert cleo.pretty_constraint == (">=0.6,<1.0" if new_format else "^0.6")
    assert not cleo.is_optional()

    pendulum = dependencies["pendulum"]
    assert pendulum.pretty_constraint == ("rev 2.0" if new_format else "branch 2.0")
    assert pendulum.is_vcs()
    pendulum = cast("VCSDependency", pendulum)
    assert pendulum.vcs == "git"
    assert pendulum.rev == "2.0" if new_format else pendulum.branch == "2.0"
    assert pendulum.source == "https://github.com/sdispater/pendulum.git"
    assert pendulum.allows_prereleases()
    assert not pendulum.develop

    tomlkit = dependencies["tomlkit"]
    assert tomlkit.pretty_constraint == "rev 3bff550"
    assert tomlkit.is_vcs()
    tomlkit = cast("VCSDependency", tomlkit)
    assert tomlkit.vcs == "git"
    assert tomlkit.rev == "3bff550"
    assert tomlkit.source == "https://github.com/sdispater/tomlkit.git"
    assert tomlkit.allows_prereleases()
    assert not tomlkit.develop if new_format else tomlkit.develop
    tomlkit_for_locking = next(d for d in package.all_requires if d.name == "tomlkit")
    assert isinstance(tomlkit_for_locking, VCSDependency)
    assert tomlkit_for_locking.develop

    requests = dependencies["requests"]
    assert requests.pretty_constraint == (
        ">=2.18,<3.0" if new_format or dynamic else "^2.18"
    )
    assert not requests.is_vcs()
    assert requests.allows_prereleases() is None
    assert requests.is_optional()
    assert requests.extras == frozenset({"security"})

    pathlib2 = dependencies["pathlib2"]
    assert pathlib2.pretty_constraint == (">=2.2,<3.0" if new_format else "^2.2")
    assert pathlib2.python_versions in {"~2.7", ">=2.7 <2.8"}
    assert not pathlib2.is_optional()

    demo = dependencies["demo"]
    assert demo.is_file()
    assert not demo.is_vcs()
    assert demo.name == "demo"
    assert demo.pretty_constraint == "*"

    demo = dependencies["my-package"]
    assert not demo.is_file()
    assert demo.is_directory()
    assert not demo.is_vcs()
    assert demo.name == "my-package"
    assert demo.pretty_constraint == "*"

    simple_project = dependencies["simple-project"]
    assert not simple_project.is_file()
    assert simple_project.is_directory()
    assert not simple_project.is_vcs()
    assert simple_project.name == "simple-project"
    assert simple_project.pretty_constraint == "*"

    functools32 = dependencies["functools32"]
    assert functools32.name == "functools32"
    assert functools32.pretty_constraint == (
        ">=3.2.3,<3.3.0" if new_format else "^3.2.3"
    )
    assert (
        str(functools32.marker)
        == 'python_version ~= "2.7" and sys_platform == "win32" or python_version in'
        ' "3.4 3.5"'
    )

    dataclasses = dependencies["dataclasses"]
    assert dataclasses.name == "dataclasses"
    assert dataclasses.pretty_constraint == (">=0.7,<1.0" if new_format else "^0.7")
    assert dataclasses.python_versions == ">=3.6.1 <3.7"
    assert (
        str(dataclasses.marker)
        == 'python_full_version >= "3.6.1" and python_version < "3.7"'
    )

    assert "db" in package.extras

    classifiers = package.classifiers

    assert classifiers == [
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]

    if new_format:
        assert package.all_classifiers == package.classifiers
    else:
        assert package.all_classifiers == [
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: 3.14",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ]


def test_create_poetry_with_empty_dependencies() -> None:
    project = "sample_project_new_no_deps"
    poetry = Factory().create_poetry(fixtures_dir / project)

    assert poetry.is_package_mode

    package = poetry.package

    assert "main" in package._dependency_groups
    assert package._dependency_groups[canonicalize_name("main")].dependencies == []


@pytest.mark.parametrize(
    "project", ["sample_project_with_groups", "sample_project_with_groups_new"]
)
def test_create_poetry_with_groups(project: str) -> None:
    poetry = Factory().create_poetry(fixtures_dir / project)

    assert "docs" in poetry.package._dependency_groups
    assert "test" in poetry.package._dependency_groups
    assert poetry.package._dependency_groups["docs"].is_optional()  # type: ignore[index]
    assert not poetry.package._dependency_groups["test"].is_optional()  # type: ignore[index]

    package = poetry.package

    expected_dependencies = {
        "test": ["pytest", "coverage"],
        "dev": ["pre-commit", "pytest", "coverage"],
        "docs": ["mkdocs"],
        "all": ["pytest", "coverage", "pre-commit", "pytest", "coverage", "mkdocs"],
    }

    dependencies = defaultdict(list)
    for dep in package.all_requires:
        assert len(dep.groups) == 1
        for group in dep.groups:
            if group != MAIN_GROUP:
                dependencies[group].append(dep.name)

    assert dependencies == expected_dependencies

    for dep in package.all_requires:
        if dep.name == "mkdocs":
            assert isinstance(dep, VCSDependency)
            # The "develop" flag of mkdocs from tool.poetry
            # must also be set for the all extra!
            assert dep.develop is True
            assert dep.source_type == "git"
            assert dep.source == "https://github.com/mkdocs/mkdocs.git"


def test_create_poetry_with_dependencies_with_subdirectory() -> None:
    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_dependencies_with_subdirectory"
    )
    package = poetry.package
    dependencies = {str(dep.name): dep for dep in package.requires}

    # git dependency
    pendulum = dependencies["pendulum"]
    assert pendulum.is_vcs()
    assert pendulum.pretty_constraint == "branch 2.0"
    pendulum = cast("VCSDependency", pendulum)
    assert pendulum.source == "https://github.com/sdispater/pendulum.git"
    assert pendulum.directory == "sub"

    # file dependency
    demo = dependencies["demo"]
    assert demo.is_file()
    assert demo.pretty_constraint == "*"
    demo = cast("FileDependency", demo)
    assert demo.path == Path("../distributions/demo-0.1.0-in-subdir.zip")
    assert demo.directory == "sub"
    demo_dependencies = [dep for dep in package.requires if dep.name == "demo"]
    assert len(demo_dependencies) == 2
    assert demo_dependencies[0] == demo_dependencies[1]
    assert {str(dep.marker) for dep in demo_dependencies} == {
        'sys_platform == "win32"',
        'sys_platform == "linux"',
    }

    # directory dependency
    simple_project = dependencies["simple-project"]
    assert simple_project.is_directory()
    assert simple_project.pretty_constraint == "*"
    simple_project = cast("DirectoryDependency", simple_project)
    assert simple_project.path == Path("../simple_project")
    with pytest.raises(AttributeError):
        _ = simple_project.directory  # type: ignore[attr-defined]

    # url dependency
    foo = dependencies["foo"]
    assert foo.is_url()
    assert foo.pretty_constraint == "*"
    foo = cast("URLDependency", foo)
    assert foo.url == "https://example.com/foo.zip"
    assert foo.directory == "sub"


def test_create_poetry_with_packages_and_includes() -> None:
    poetry = Factory().create_poetry(
        fixtures_dir.parent / "masonry" / "builders" / "fixtures" / "with-include"
    )

    package = poetry.package

    assert package.packages == [
        {"include": "extra_dir/**/*.py", "format": ["sdist", "wheel"]},
        {"include": "extra_dir/**/*.py", "format": ["sdist", "wheel"]},
        {"include": "my_module.py", "format": ["sdist", "wheel"]},
        {"include": "package_with_include", "format": ["sdist", "wheel"]},
        {"include": "tests", "format": ["sdist"]},
        {"include": "for_wheel_only", "format": ["wheel"]},
        {"include": "src_package", "from": "src", "format": ["sdist", "wheel"]},
        {
            "include": "from_to",
            "from": "etc",
            "to": "target_from_to",
            "format": ["sdist", "wheel"],
        },
        {
            "include": "my_module_to.py",
            "to": "target_module",
            "format": ["sdist", "wheel"],
        },
    ]

    assert package.include == [
        {"path": "extra_dir/vcs_excluded.py", "format": ["sdist", "wheel"]},
        {"path": "notes.txt", "format": ["sdist"]},
    ]


def test_create_poetry_with_multi_constraints_dependency() -> None:
    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_multi_constraints_dependency"
    )

    package = poetry.package

    assert len(package.requires) == 2


def test_create_poetry_non_package_mode() -> None:
    poetry = Factory().create_poetry(fixtures_dir / "non_package_mode")

    assert not poetry.is_package_mode


@pytest.mark.parametrize(
    "project", ["none", "file", "text", "text_spdx", "str", "str_empty", "str_no_spdx"]
)
@pytest.mark.parametrize("with_license_files", [False, True, "empty"])
def test_create_poetry_with_license_type(
    project: str, with_license_files: bool | str, tmp_path: Path
) -> None:
    project_dir = fixtures_dir / f"with_license_type_{project}"
    expected_license_files: tuple[str, ...] | Path | None = None
    if with_license_files:
        if with_license_files == "empty":
            content = ""
            expected_license_files = ()
        else:
            content = '"LICEN[CS]E*", "AUTHORS*"'
            expected_license_files = ("LICEN[CS]E*", "AUTHORS*")

        orig_project_dir = project_dir
        project_dir = tmp_path / project
        shutil.copytree(orig_project_dir, project_dir)
        pyproject_file = project_dir / "pyproject.toml"
        new_lines = []
        for line in pyproject_file.read_text(encoding="utf-8").splitlines():
            if line.startswith("keywords = "):
                new_lines.append(f"license-files = [{content}]")
            new_lines.append(line)
        pyproject_file.write_text("\n".join(new_lines), encoding="utf-8")

    license_type = project.split("_", 1)[0]
    expected_license_id: str | None = None
    expected_license_expression: str | None = None
    if license_type == "none":
        pass
    elif license_type == "file":
        expected_license_id = (project_dir / "LICENSE").read_text(encoding="utf-8")
        expected_license_files = Path("LICENSE")
    elif license_type in {"str", "text"}:
        with (project_dir / "pyproject.toml").open("rb") as f:
            data = tomllib.load(f)
        project_license = data["project"]["license"]
        if license_type == "text":
            expected_license_id = project_license["text"]
        elif project == "str_no_spdx":
            expected_license_id = project_license
        elif project == "str":
            expected_license_expression = project_license
    else:
        raise RuntimeError("unexpected license type")

    if with_license_files and license_type in {"file", "text"}:
        with pytest.raises(ValueError) as e:
            Factory().create_poetry(project_dir)
        assert str(e.value) == (
            "[project.license] must be of type string"
            " if [project.license-files] is defined."
        )
    else:
        poetry = Factory().create_poetry(project_dir)

        if expected_license_id is None:
            assert poetry.package.license is None
        else:
            assert poetry.package.license is not None
            assert poetry.package.license.id == expected_license_id
        assert poetry.package.license_expression == expected_license_expression
        assert poetry.package.license_files == expected_license_files


@pytest.mark.parametrize(
    ("invalid_glob", "expected_message"),
    [
        (
            r"sub\\LICENSE",
            (
                "Invalid entry in [project.license-files]: 'sub\\LICENSE'"
                " (Path delimiters must be forward slashes.)"
            ),
        ),
        (
            "../LICENSE",
            (
                "Invalid entry in [project.license-files]: '../LICENSE'"
                " ('..' must not be used.)"
            ),
        ),
        (
            "./../LICENSE",
            (
                "Invalid entry in [project.license-files]: './../LICENSE'"
                " ('..' must not be used.)"
            ),
        ),
        (
            "sub/../../LICENSE",
            (
                "Invalid entry in [project.license-files]: 'sub/../../LICENSE'"
                " ('..' must not be used.)"
            ),
        ),
    ],
)
def test_create_poetry_with_invalid_license_files_glob(
    tmp_path: Path, invalid_glob: str, expected_message: str
) -> None:
    project_file = tmp_path / "pyproject.toml"
    project_file.write_text(
        f"""\
[project]
name = "foo"
version = "1"
license-files = [
    "LICENSE",
    "{invalid_glob}",
    "licenses/**",
]
""",
        encoding="utf-8",
    )
    with pytest.raises(ValueError) as e:
        Factory().create_poetry(tmp_path)
    assert str(e.value) == expected_message


def test_create_poetry_fails_with_missing_license_file() -> None:
    project_dir = fixtures_dir / "missing_license_file"
    with pytest.raises(FileNotFoundError) as e:
        Factory().create_poetry(project_dir)

    assert str((project_dir / "LICENSE").absolute()) in str(e.value)


@pytest.mark.parametrize(
    ("requires_python", "python", "expected_versions", "expected_constraint"),
    [
        (">=3.8", None, ">=3.8", ">=3.8"),
        (None, "^3.8", "^3.8", ">=3.8,<4.0"),
        (">=3.8", "^3.8", "^3.8", ">=3.8,<4.0"),
    ],
)
def test_create_poetry_python_version(
    requires_python: str,
    python: str,
    expected_versions: str,
    expected_constraint: str,
    tmp_path: Path,
) -> None:
    content = '[project]\nname = "foo"\nversion = "1"\n'
    if requires_python:
        content += f'requires-python = "{requires_python}"\n'
    if python:
        content += f'[tool.poetry.dependencies]\npython = "{python}"\n'
    (tmp_path / "pyproject.toml").write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(tmp_path)

    package = poetry.package
    assert package.requires_python == requires_python or python
    assert package.python_versions == expected_versions
    assert str(package.python_constraint) == expected_constraint


def test_create_poetry_python_version_not_compatible(tmp_path: Path) -> None:
    content = """
[project]
name = "foo"
version = "1"
requires-python = ">=3.8"

[tool.poetry.dependencies]
python = ">=3.7"
"""
    (tmp_path / "pyproject.toml").write_text(content, encoding="utf-8")
    with pytest.raises(ValueError) as e:
        Factory().create_poetry(tmp_path)

    assert "not a subset" in str(e.value)


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (  # static
            """\
[project]
name = "foo"
version = "1"
requires-python = "3.10"
classifiers = ["License :: OSI Approved :: MIT License"]
""",
            ["License :: OSI Approved :: MIT License"],
        ),
        (  # dynamic
            """\
[project]
name = "foo"
version = "1"
requires-python = "3.10"
dynamic = [ "classifiers" ]

[tool.poetry]
classifiers = ["License :: OSI Approved :: MIT License"]
""",
            [
                "License :: OSI Approved :: MIT License",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.10",
            ],
        ),
        (  # legacy
            """\
[tool.poetry]
name = "foo"
version = "1"
classifiers = ["License :: OSI Approved :: MIT License"]

[tool.poetry.dependencies]
python = "~3.10"
""",
            [
                "License :: OSI Approved :: MIT License",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.10",
            ],
        ),
    ],
)
def test_create_poetry_classifiers(
    content: str, expected: list[str], tmp_path: Path
) -> None:
    (tmp_path / "pyproject.toml").write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(tmp_path)

    assert poetry.package.all_classifiers == expected


def test_create_poetry_no_readme(tmp_path: Path) -> None:
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(
        '[tool.poetry]\nname="foo"\nversion="1"\nauthors = []\ndescription = ""\n',
        encoding="utf-8",
    )
    poetry = Factory().create_poetry(tmp_path)

    assert not poetry.package.readmes


def test_create_poetry_empty_readme(tmp_path: Path) -> None:
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(
        '[tool.poetry]\nname="foo"\nversion="1"\nauthors = []\ndescription = ""\n'
        'readme = ""\n',
        encoding="utf-8",
    )
    poetry = Factory().create_poetry(tmp_path)

    assert not poetry.package.readmes


def test_validate() -> None:
    complete = fixtures_dir / "complete.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content) == {"errors": [], "warnings": []}


def test_validate_strict_legacy_warnings(complete_legacy_warnings: list[str]) -> None:
    complete = fixtures_dir / "complete.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {
        "errors": [],
        "warnings": complete_legacy_warnings,
    }


def test_validate_strict_legacy_duplicate_warnings(
    complete_legacy_duplicate_warnings: list[str],
) -> None:
    complete = fixtures_dir / "complete_duplicates.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {
        "errors": [],
        "warnings": complete_legacy_duplicate_warnings,
    }


def test_validate_strict_new_no_warnings() -> None:
    complete = fixtures_dir / "complete_new.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {"errors": [], "warnings": []}


def test_validate_strict_dynamic_warnings() -> None:
    # some fields are allowed to be dynamic, but some are not
    complete = fixtures_dir / "complete_new_dynamic_invalid.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {
        "errors": ["project must contain ['name'] properties"],
        "warnings": [
            # version, readme and classifiers are allowed to be dynamic!
            "[tool.poetry.name] is deprecated. Use [project.name] instead.",
            (
                "[tool.poetry.description] is deprecated. Use "
                "[project.description] instead."
            ),
            "[tool.poetry.license] is deprecated. Use [project.license] instead.",
            "[tool.poetry.authors] is deprecated. Use [project.authors] instead.",
            (
                "[tool.poetry.maintainers] is deprecated. Use "
                "[project.maintainers] instead."
            ),
            "[tool.poetry.keywords] is deprecated. Use [project.keywords] instead.",
            "[tool.poetry.homepage] is deprecated. Use [project.urls] instead.",
            "[tool.poetry.repository] is deprecated. Use [project.urls] instead.",
            "[tool.poetry.documentation] is deprecated. Use [project.urls] instead.",
            (
                "[tool.poetry.extras] is deprecated. Use "
                "[project.optional-dependencies] instead."
            ),
            (
                "Defining console scripts in [tool.poetry.scripts] is deprecated. "
                "Use [project.scripts] instead. "
                "([tool.poetry.scripts] should only be used for scripts of type 'file')."
            ),
        ],
    }


def test_validate_local_version(tmp_path: Path) -> None:
    project = tmp_path / "local_version.toml"
    project.write_text(
        """[project]\nname = "local-version"\nversion = "0.5.0+LOCAL.123A"\n""",
        encoding="utf-8",
    )
    with project.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content) == {"errors": [], "warnings": []}


def test_validate_fails() -> None:
    complete = fixtures_dir / "complete.toml"
    with complete.open("rb") as f:
        content = tomllib.load(f)
    content["tool"]["poetry"]["authors"] = "this is not a valid array"

    expected = "tool.poetry.authors must be array"

    assert Factory.validate(content) == {"errors": [expected], "warnings": []}


def test_validate_without_strict_fails_only_non_strict() -> None:
    project_failing_strict_validation = (
        fixtures_dir / "project_failing_strict_validation" / "pyproject.toml"
    )
    with project_failing_strict_validation.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content) == {
        "errors": [
            "Either [project.name] or [tool.poetry.name] is required in package mode.",
            (
                "Either [project.version] or [tool.poetry.version] is required in "
                "package mode."
            ),
        ],
        "warnings": [],
    }


def test_validate_strict_fails_strict_and_non_strict() -> None:
    project_failing_strict_validation = (
        fixtures_dir / "project_failing_strict_validation" / "pyproject.toml"
    )
    with project_failing_strict_validation.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {
        "errors": [
            "Either [project.name] or [tool.poetry.name] is required in package mode.",
            (
                "Either [project.version] or [tool.poetry.version] is required in "
                "package mode."
            ),
            (
                'Cannot find dependency "missing_extra" for extra "some-extras" in '
                "main dependencies."
            ),
            (
                'Cannot find dependency "another_missing_extra" for extra '
                '"some-extras" in main dependencies.'
            ),
            (
                'The script "a_script_with_unknown_extra" requires extra "foo" which is'
                " not defined."
            ),
            (
                "Declared README files must be of same type: found text/markdown,"
                " text/x-rst"
            ),
        ],
        "warnings": [
            (
                "[tool.poetry.readme] is set but 'readme' is not in "
                "[project.dynamic]. If it is static use [project.readme]. If it "
                "is dynamic, add 'readme' to [project.dynamic].\n"
                "If you want to define multiple readmes, you should define them "
                "in [tool.poetry] and add 'readme' to [project.dynamic]."
            ),
            (
                "[tool.poetry.extras] is deprecated. Use "
                "[project.optional-dependencies] instead."
            ),
            (
                "Defining console scripts in [tool.poetry.scripts] is deprecated. "
                "Use [project.scripts] instead. "
                "([tool.poetry.scripts] should only be used for scripts of type 'file')."
            ),
            (
                "A wildcard Python dependency is ambiguous. Consider specifying a more"
                " explicit one."
            ),
            (
                'The "pathlib2" dependency specifies the "allows-prereleases" property,'
                ' which is deprecated. Use "allow-prereleases" instead.'
            ),
            (
                'The script "a_script_with_unknown_extra" depends on an extra. Scripts'
                " depending on extras are deprecated and support for them will be"
                " removed in a future version of poetry/poetry-core. See"
                " https://packaging.python.org/en/latest/specifications/entry-points/#data-model"
                " for details."
            ),
        ],
    }


@pytest.mark.parametrize("with_project_section", [True, False])
def test_validate_dependencies_non_package_mode(with_project_section: bool) -> None:
    content: dict[str, Any] = {
        "tool": {"poetry": {"package-mode": False, "dependencies": {"foo": "*"}}}
    }
    expected: dict[str, list[str]] = {"errors": [], "warnings": []}
    if with_project_section:
        content["project"] = {"name": "my-project"}
        expected["warnings"] = [
            (
                "[tool.poetry.dependencies] is set but [project.dependencies] is "
                "not and 'dependencies' is not in [project.dynamic]. You should "
                "either migrate [tool.poetry.dependencies] to "
                "[project.dependencies] (if you do not need Poetry-specific "
                "features) or add [project.dependencies] in addition to "
                "[tool.poetry.dependencies] or add 'dependencies' to "
                "[project.dynamic]."
            )
        ]
    assert Factory.validate(content, strict=True) == expected


@pytest.mark.parametrize("with_project_section", [True, False])
def test_validate_python_non_package_mode(with_project_section: bool) -> None:
    content: dict[str, Any] = {
        "tool": {"poetry": {"package-mode": False, "dependencies": {"python": ">=3.9"}}}
    }
    expected: dict[str, list[str]] = {"errors": [], "warnings": []}
    if with_project_section:
        content["project"] = {"name": "my-project", "dynamic": ["dependencies"]}
        expected["warnings"] = [
            (
                "[tool.poetry.dependencies.python] is set but [project.requires-python]"
                " is not set and 'requires-python' is not in [project.dynamic]."
            )
        ]
    assert Factory.validate(content, strict=True) == expected


@pytest.mark.parametrize("section", ["project", "poetry"])
@pytest.mark.parametrize("with_license_classifier", [True, False])
def test_validate_deprecated_license_classifiers(
    section: str, with_license_classifier: bool
) -> None:
    content: dict[str, Any] = {
        "project": {"name": "my-project", "version": "1.0", "license": "MIT"},
        "tool": {"poetry": {}},
    }
    classifiers = ["Topic :: Software Development :: Libraries :: Python Modules"]

    expected: dict[str, list[str]] = {"errors": [], "warnings": []}
    if with_license_classifier:
        classifiers.append("License :: OSI Approved :: MIT License")
        expected["warnings"].append(
            "License classifiers are deprecated. Use [project.license] instead."
        )

    if section == "project":
        content["project"]["classifiers"] = classifiers
    elif section == "poetry":
        content["tool"]["poetry"]["classifiers"] = classifiers
        content["project"]["dynamic"] = ["classifiers"]
    else:
        raise RuntimeError("unexpected section")

    assert Factory.validate(content, strict=True) == expected


@pytest.mark.parametrize(
    "project", ["none", "file", "text", "text_spdx", "str", "str_empty", "str_no_spdx"]
)
@pytest.mark.parametrize("with_license_files", [False, True])
def test_validate_with_license_type(
    project: str, with_license_files: bool, tmp_path: Path
) -> None:
    project_dir = fixtures_dir / f"with_license_type_{project}"
    pyproject_file = project_dir / "pyproject.toml"
    if with_license_files:
        orig_project_dir = project_dir
        project_dir = tmp_path / project
        shutil.copytree(orig_project_dir, project_dir)
        pyproject_file = project_dir / "pyproject.toml"
        new_lines = []
        for line in pyproject_file.read_text(encoding="utf-8").splitlines():
            if line.startswith("keywords = "):
                new_lines.append('license-files = ["LICEN[CS]E*", "AUTHORS*"]')
            new_lines.append(line)
        pyproject_file.write_text("\n".join(new_lines), encoding="utf-8")

    expected_warnings = []
    if project.split("_", 1)[0] in {"file", "text"}:
        expected_warnings.append(
            "Defining [project.license] as a table is deprecated."
            " [project.license] should be a valid SPDX license expression."
            " License files can be referenced in [project.license-files]."
        )
    elif project in {"str_empty", "str_no_spdx"}:
        expected_warnings.append(
            "[project.license] is not a valid SPDX expression."
            " This is deprecated and will raise an error in the future."
        )

    with pyproject_file.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {
        "errors": [],
        "warnings": expected_warnings,
    }


def test_strict_validation_success_on_multiple_readme_files() -> None:
    with_readme_files = fixtures_dir / "with_readme_files" / "pyproject.toml"
    with with_readme_files.open("rb") as f:
        content = tomllib.load(f)

    assert Factory.validate(content, strict=True) == {"errors": [], "warnings": []}


def test_strict_validation_fails_on_readme_files_with_unmatching_types() -> None:
    with_readme_files = fixtures_dir / "with_readme_files" / "pyproject.toml"
    with with_readme_files.open("rb") as f:
        content = tomllib.load(f)
    content["tool"]["poetry"]["readme"][0] = "README.md"

    assert Factory.validate(content, strict=True) == {
        "errors": [
            "Declared README files must be of same type: found text/markdown,"
            " text/x-rst"
        ],
        "warnings": [],
    }


def test_create_poetry_fails_on_invalid_configuration() -> None:
    with pytest.raises(RuntimeError) as e:
        Factory().create_poetry(
            Path(__file__).parent / "fixtures" / "invalid_pyproject" / "pyproject.toml"
        )

    expected = """\
The Poetry configuration is invalid:
  - Either [project.name] or [tool.poetry.name] is required in package mode.
  - Either [project.version] or [tool.poetry.version] is required in package mode.
"""
    assert str(e.value) == expected


def test_create_poetry_fails_on_invalid_mode() -> None:
    with pytest.raises(RuntimeError) as e:
        Factory().create_poetry(
            Path(__file__).parent / "fixtures" / "invalid_mode" / "pyproject.toml"
        )

    expected = """\
The Poetry configuration is invalid:
  - tool.poetry.package-mode must be boolean
  - Either [project.name] or [tool.poetry.name] is required in package mode.
  - Either [project.version] or [tool.poetry.version] is required in package mode.
"""
    assert str(e.value) == expected


def test_create_poetry_omits_dev_dependencies_iff_with_dev_is_false() -> None:
    poetry = Factory().create_poetry(fixtures_dir / "sample_project", with_groups=False)
    assert not any("dev" in r.groups for r in poetry.package.all_requires)

    poetry = Factory().create_poetry(fixtures_dir / "sample_project")
    assert any("dev" in r.groups for r in poetry.package.all_requires)


def test_create_poetry_with_invalid_dev_dependencies(caplog: LogCaptureFixture) -> None:
    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_invalid_dev_deps", with_groups=False
    )
    assert not any("dev" in r.groups for r in poetry.package.all_requires)

    assert not caplog.records
    poetry = Factory().create_poetry(fixtures_dir / "project_with_invalid_dev_deps")
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "does not exist" in record.message
    assert any("dev" in r.groups for r in poetry.package.all_requires)


@pytest.mark.parametrize("with_groups", [True, False])
def test_create_poetry_with_invalid_dependency_groups(with_groups: bool) -> None:
    with pytest.raises(RuntimeError) as e:
        _ = Factory().create_poetry(
            fixtures_dir / "project_with_invalid_dependency_groups",
            with_groups=with_groups,
        )

    expected = """\
The Poetry configuration is invalid:
  - dependency-groups.testing[1] must be valid exactly by one definition\
 (0 matches found)
"""
    assert str(e.value) == expected


def test_create_poetry_with_duplicated_dependency_groups() -> None:
    with pytest.raises(RuntimeError) as e:
        _ = Factory().create_poetry(
            fixtures_dir / "project_with_duplicated_dependency_groups",
        )

    assert (
        "Duplicate dependency group name after normalization: test (Test, test)"
        in str(e.value)
    )


def test_create_poetry_with_dependency_groups_missing_include() -> None:
    with pytest.raises(ValueError) as e:
        _ = Factory().create_poetry(
            fixtures_dir / "project_with_dependency_groups_missing_include",
        )

    assert (
        str(e.value) == "Group 'test' includes group 'coverage' which is not defined."
    )


@pytest.mark.parametrize(
    ("fixture", "expected"),
    [
        (
            "project_with_dependency_groups_simple_cycle",
            [
                "Cyclic dependency group include in test: coverage -> test",
                "Cyclic dependency group include in coverage: test -> coverage",
            ],
        ),
        (
            "project_with_dependency_groups_complex_cycle",
            [
                "Cyclic dependency group include in test: coverage -> dev -> test",
                "Cyclic dependency group include in coverage: dev -> test -> coverage",
                "Cyclic dependency group include in dev: test -> coverage -> dev",
            ],
        ),
    ],
)
def test_create_poetry_with_dependency_groups_simple_cycle(
    fixture: str, expected: list[str]
) -> None:
    with pytest.raises(RuntimeError) as e:
        _ = Factory().create_poetry(fixtures_dir / fixture)

    assert all(exp in str(e.value) for exp in expected)


def test_create_poetry_with_groups_and_legacy_dev(caplog: LogCaptureFixture) -> None:
    assert not caplog.records

    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_groups_and_legacy_dev"
    )

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert '"poetry.dev-dependencies" section is deprecated' in record.message

    package = poetry.package
    dependencies = package.all_requires

    assert len(dependencies) == 2
    assert {dependency.name for dependency in dependencies} == {"pytest", "pre-commit"}


def test_create_poetry_with_groups_and_explicit_main() -> None:
    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_groups_and_explicit_main"
    )

    package = poetry.package
    dependencies = package.requires

    assert len(dependencies) == 1
    assert {dependency.name for dependency in dependencies} == {
        "aiohttp",
    }


def test_create_poetry_with_markers_and_extras() -> None:
    poetry = Factory().create_poetry(fixtures_dir / "project_with_markers_and_extras")

    package = poetry.package
    dependencies = package.requires
    extras = package.extras

    assert len(dependencies) == 2
    assert {dependency.name for dependency in dependencies} == {"orjson"}
    assert set(extras[canonicalize_name("all")]) == set(dependencies)
    for dependency in dependencies:
        assert dependency.in_extras == ["all"]
        assert isinstance(dependency, URLDependency)
        assert isinstance(dependency.marker, SingleMarker)
        assert dependency.marker.name == "sys_platform"
        assert dependency.marker.value == (
            "darwin" if "macosx" in dependency.url else "linux"
        )


@pytest.mark.parametrize(
    "constraint, exp_python, exp_marker",
    [
        ({"python": "3.7"}, "~3.7", 'python_version == "3.7"'),
        ({"platform": "linux"}, "*", 'sys_platform == "linux"'),
        ({"markers": 'python_version == "3.7"'}, "~3.7", 'python_version == "3.7"'),
        (
            {"markers": 'platform_machine == "x86_64"'},
            "*",
            'platform_machine == "x86_64"',
        ),
        (
            {"python": "3.7", "markers": 'platform_machine == "x86_64"'},
            "~3.7",
            'platform_machine == "x86_64" and python_version == "3.7"',
        ),
        (
            {"platform": "linux", "markers": 'platform_machine == "x86_64"'},
            "*",
            'platform_machine == "x86_64" and sys_platform == "linux"',
        ),
        (
            {
                "python": "3.7",
                "platform": "linux",
                "markers": 'platform_machine == "x86_64"',
            },
            "~3.7",
            (
                'platform_machine == "x86_64" and python_version == "3.7" and'
                ' sys_platform == "linux"'
            ),
        ),
        (
            {"python": ">=3.7", "markers": 'python_version < "4.0"'},
            "<4.0 >=3.7",
            'python_version < "4.0" and python_version >= "3.7"',
        ),
        (
            {"platform": "linux", "markers": 'sys_platform == "win32"'},
            "*",
            "<empty>",
        ),
    ],
)
def test_create_dependency_marker_variants(
    constraint: dict[str, Any], exp_python: str, exp_marker: str
) -> None:
    constraint["version"] = "1.0.0"
    dep = Factory.create_dependency("foo", constraint)
    assert dep.python_versions == exp_python
    assert dep.python_constraint == parse_constraint(exp_python)
    assert str(dep.marker) == exp_marker


@pytest.mark.parametrize(
    ("constraint", "expected"),
    [
        ("1", None),
        ({"version": "1"}, None),
        ({"version": "1", "allow-prereleases": False}, False),
        ({"version": "1", "allow-prereleases": True}, True),
    ],
)
def test_create_dependency_allow_prereleases(
    constraint: str | dict[str, str], expected: bool | None
) -> None:
    dep = Factory.create_dependency("foo", constraint)
    assert dep.allows_prereleases() is expected


def test_all_classifiers_unique_even_if_classifiers_is_duplicated() -> None:
    poetry = Factory().create_poetry(
        fixtures_dir / "project_with_duplicated_classifiers"
    )
    package = poetry.package
    assert package.all_classifiers == [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Topic :: Software Development :: Build Tools",
    ]


@pytest.mark.parametrize(
    ("project", "expected"),
    [
        ("sample_project", set(BuildSystem().dependencies)),
        (
            "project_with_build_system_requires",
            {
                Dependency.create_from_pep_508("poetry-core"),
                Dependency.create_from_pep_508("Cython (>=0.29.6,<0.30.0)"),
            },
        ),
    ],
)
def test_poetry_build_system_dependencies_from_fixtures(
    project: str, expected: set[Dependency]
) -> None:
    poetry = Factory().create_poetry(fixtures_dir / project)
    assert set(poetry.build_system_dependencies) == expected


SAMPLE_PROJECT_DIRECTORY = fixtures_dir / "sample_project"
SIMPLE_PROJECT_WHEEL = (
    fixtures_dir
    / "simple_project"
    / "dist"
    / "simple_project-1.2.3-py2.py3-none-any.whl"
)


@pytest.mark.parametrize(
    ("requires", "expected"),
    [
        (BuildSystem().requires, set(BuildSystem().dependencies)),
        (["poetry-core>=2.0.0"], {Dependency("poetry-core", ">=2.0.0")}),
        (["****invalid****"], set()),
        (
            ["hatch", "numpy ; sys_platform == 'win32'"],
            {
                Dependency("hatch", "*"),
                Dependency.create_from_pep_508("numpy ; sys_platform == 'win32'"),
            },
        ),
        (
            [SAMPLE_PROJECT_DIRECTORY.as_posix()],
            {
                DirectoryDependency(
                    SAMPLE_PROJECT_DIRECTORY.name, SAMPLE_PROJECT_DIRECTORY
                )
            },
        ),
        (
            [SIMPLE_PROJECT_WHEEL.as_posix()],
            {FileDependency(SIMPLE_PROJECT_WHEEL.name, SIMPLE_PROJECT_WHEEL)},
        ),
    ],
)
def test_poetry_build_system_dependencies(
    requires: list[str], expected: set[Dependency], temporary_directory: Path
) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"
    build_system_requires = ", ".join(f'"{require}"' for require in requires)
    content = f"""[project]
name = "my-package"
version = "1.2.3"

[build-system]
requires = [{build_system_requires}]
build-backend = "some.api.we.do.not.care.about"

"""
    pyproject_toml.write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(temporary_directory)

    assert set(poetry.build_system_dependencies) == expected


@pytest.mark.parametrize("in_order", [True, False])
@pytest.mark.parametrize(
    ("group_name", "included_group_name"),
    [
        ("testing", "testing"),
        ("testing", "TESTING"),
        ("group_a", "group-a"),
        # Examples from the PEP 508 spec
        # https://packaging.python.org/en/latest/specifications/name-normalization/#valid-non-normalized-names
        ("friendly-bard", "friendly-bard"),
        ("friendly-bard", "Friendly-Bard"),
        ("friendly-bard", "FRIENDLY-BARD"),
        ("friendly-bard", "friendly.bard"),
        ("friendly-bard", "friendly_bard"),
        ("friendly-bard", "friendly--bard"),
        ("friendly-bard", "FrIeNdLy-._.-bArD"),
        ("friendly-Bard", "friendly-bard"),
        ("FRIENDLY-BARD", "friendly-bard"),
        ("friendly.bard", "friendly-bard"),
        ("friendly_bard", "friendly-bard"),
        ("friendly--bard", "friendly-bard"),
        ("FrIeNdLy-._.-bArD", "friendly-bard"),
    ],
)
def test_create_poetry_with_nested_dependency_groups(
    group_name: str, included_group_name: str, in_order: bool, temporary_directory: Path
) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"

    replace_group_name = "%REPLACE_GROUP_NAME%"
    replace_included_group_name = "%REPLACE_INCLUDED_GROUP_NAME%"
    in_order_content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.%REPLACE_GROUP_NAME%.dependencies]
pytest = "*"
pytest-cov ="*"

[tool.poetry.group.dev]
include-groups = [
    "%REPLACE_INCLUDED_GROUP_NAME%",
]
[tool.poetry.group.dev.dependencies]
black = "*"
"""
    # The dev group refers to a group that is defined after it.
    out_of_order_content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.dev]
include-groups = [
    "%REPLACE_INCLUDED_GROUP_NAME%",
]
[tool.poetry.group.dev.dependencies]
black = "*"

[tool.poetry.group.%REPLACE_GROUP_NAME%.dependencies]
pytest = "*"
pytest-cov ="*"
"""

    # Generate the content. If `group_name` has a `.` in it, we "escape" it with
    # quotes to make it a valid TOML key.
    base_content = in_order_content if in_order else out_of_order_content
    group_name_to_use = group_name if "." not in group_name else f'"{group_name}"'
    content = base_content.replace(replace_group_name, group_name_to_use).replace(
        replace_included_group_name, included_group_name
    )

    pyproject_toml.write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(temporary_directory)

    # Groups are reported internally using their canonical names.
    canonical_name = canonicalize_name(group_name)

    assert len(poetry.package.all_requires) == 5
    assert sorted(
        [(dep.name, ",".join(dep.groups)) for dep in poetry.package.all_requires],
        key=lambda x: x[0] + x[1],
    ) == sorted(
        [
            ("black", "dev"),
            ("pytest-cov", "dev"),
            ("pytest-cov", canonical_name),
            ("pytest", "dev"),
            ("pytest", canonical_name),
        ],
        key=lambda x: x[0] + x[1],
    )


def assert_invalid_group_including(
    toml_data: str,
    expected_error: str,
    error_type: type[Exception],
    temporary_directory: Path,
) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"
    pyproject_toml.write_text(toml_data, encoding="utf-8")

    with pytest.raises(error_type) as error:
        _ = Factory().create_poetry(temporary_directory)

    assert str(error.value) == expected_error


@pytest.mark.parametrize(
    "include_group_name", ["testing_group", "Testing-Group", "testing-group"]
)
def test_create_poetry_with_self_referenced_dependency_groups(
    include_group_name: str,
    temporary_directory: Path,
) -> None:
    """testing-group -> testing-group"""
    content = f"""\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.testing-group]
include-groups = [
    "{include_group_name}",
]

[tool.poetry.group.testing-group.dependencies]
pytest = "*"
pytest-cov ="*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in testing-group: testing-group
"""
    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_direct_cyclic_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    testing -> dev
    dev -> testing
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.testing]
include-groups = [
    "dev",
]

[tool.poetry.group.testing.dependencies]
pytest = "*"
pytest-cov ="*"

[tool.poetry.group.dev]
include-groups = [
    "testing",
]
[tool.poetry.group.dev.dependencies]
black = "*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in testing: dev -> testing
  - Cyclic dependency group include in dev: testing -> dev
"""
    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_indirect_full_cyclic_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    group-1 -> group-3
    group-2 -> group-1
    group-3 -> group-2
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.group_1]
include-groups = [
    "group_3",
]

[tool.poetry.group.group_1.dependencies]
foo = "*"

[tool.poetry.group.group_2]
include-groups = [
    "group_1",
]
[tool.poetry.group.group_2.dependencies]
bar = "*"

[tool.poetry.group.group_3]
include-groups = [
    "group_2",
]
[tool.poetry.group.group_3.dependencies]
baz = "*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in group-1: group-3 -> group-2 -> group-1
  - Cyclic dependency group include in group-2: group-1 -> group-3 -> group-2
  - Cyclic dependency group include in group-3: group-2 -> group-1 -> group-3
"""
    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_indirect_partial_cyclic_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    group-1 -> group-2
    group-2 -> group-1
    group-3 -> group-2
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.group_1]
include-groups = [
    "group_2",
]

[tool.poetry.group.group_1.dependencies]
foo = "*"

[tool.poetry.group.group_2]
include-groups = [
    "group_1",
]
[tool.poetry.group.group_2.dependencies]
bar = "*"

[tool.poetry.group.group_3]
include-groups = [
    "group_2",
]
[tool.poetry.group.group_3.dependencies]
baz = "*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in group-1: group-2 -> group-1
  - Cyclic dependency group include in group-2: group-1 -> group-2
  - Cyclic dependency group include in group-3: group-2 -> group-1 -> group-2
"""
    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_shared_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    root -> child-1, child-2
    child-1 -> shared
    child-2 -> shared
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.root]
include-groups = [
    "child_1",
    "child_2",
]

[tool.poetry.group.root.dependencies]
foo = "*"

[tool.poetry.group.child_1]
include-groups = [
    "shared",
]
[tool.poetry.group.child_1.dependencies]
bar = "*"

[tool.poetry.group.child_2]
include-groups = [
    "shared",
]
[tool.poetry.group.child_2.dependencies]
baz = "*"

[tool.poetry.group.shared.dependencies]
quux = "*"
"""
    pyproject_toml = temporary_directory / "pyproject.toml"
    pyproject_toml.write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(temporary_directory)

    assert len(poetry.package.all_requires) == 10
    assert sorted(
        [(dep.name, ",".join(dep.groups)) for dep in poetry.package.all_requires],
        key=lambda x: x[0] + x[1],
    ) == [
        ("bar", "child-1"),
        ("bar", "root"),
        ("baz", "child-2"),
        ("baz", "root"),
        ("foo", "root"),
        ("quux", "child-1"),
        ("quux", "child-2"),
        # Duplicates because dependency is included via several groups.
        # This is ok because they are merged during dependency resolution.
        ("quux", "root"),
        ("quux", "root"),
        ("quux", "shared"),
    ]


def test_create_poetry_with_shared_dependency_groups_more_complicated(
    temporary_directory: Path,
) -> None:
    """
    root -> child-1, child-2
    child-1 -> shared
    child-2 -> grandchild
    grandchild -> shared
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.root]
include-groups = [
    "child_1",
    "child_2",
]

[tool.poetry.group.root.dependencies]
foo = "*"

[tool.poetry.group.child_1]
include-groups = [
    "shared",
]
[tool.poetry.group.child_1.dependencies]
bar = "*"

[tool.poetry.group.child_2]
include-groups = [
    "grandchild",
]
[tool.poetry.group.child_2.dependencies]
baz = "*"

[tool.poetry.group.grandchild]
include-groups = [
    "shared",
]
[tool.poetry.group.grandchild.dependencies]
bax = "*"

[tool.poetry.group.shared.dependencies]
quux = "*"
"""
    pyproject_toml = temporary_directory / "pyproject.toml"
    pyproject_toml.write_text(content, encoding="utf-8")
    poetry = Factory().create_poetry(temporary_directory)

    assert len(poetry.package.all_requires) == 14
    assert sorted(
        [(dep.name, ",".join(dep.groups)) for dep in poetry.package.all_requires],
        key=lambda x: x[0] + x[1],
    ) == [
        ("bar", "child-1"),
        ("bar", "root"),
        ("bax", "child-2"),
        ("bax", "grandchild"),
        ("bax", "root"),
        ("baz", "child-2"),
        ("baz", "root"),
        ("foo", "root"),
        ("quux", "child-1"),
        ("quux", "child-2"),
        ("quux", "grandchild"),
        # Duplicates because dependency is included via several groups.
        # This is ok because they are merged during dependency resolution.
        ("quux", "root"),
        ("quux", "root"),
        ("quux", "shared"),
    ]


def test_create_poetry_with_complicated_cyclic_diamond_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    root -> child-1, child-2
    child-1 -> shared
    child-2 -> shared
    shared -> grandchild
    grandchild -> child-2
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.root]
include-groups = [
    "child_1",
    "child_2",
]

[tool.poetry.group.root.dependencies]
foo = "*"

[tool.poetry.group.child_1]
include-groups = [
    "shared",
]
[tool.poetry.group.child_1.dependencies]
bar = "*"

[tool.poetry.group.child_2]
include-groups = [
    "shared",
]
[tool.poetry.group.child_2.dependencies]
baz = "*"

[tool.poetry.group.shared]
include-groups = [
    "grandchild",
]
[tool.poetry.group.shared.dependencies]
quux = "*"

[tool.poetry.group.grandchild]
include-groups = [
    "child_2",
]
[tool.poetry.group.grandchild.dependencies]
bar = "*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in root: child-2 -> shared -> grandchild -> child-2
  - Cyclic dependency group include in root: child-1 -> shared -> grandchild -> child-2 -> shared
  - Cyclic dependency group include in child-1: shared -> grandchild -> child-2 -> shared
  - Cyclic dependency group include in child-2: shared -> grandchild -> child-2
  - Cyclic dependency group include in shared: grandchild -> child-2 -> shared
  - Cyclic dependency group include in grandchild: child-2 -> shared -> grandchild
"""

    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_noncanonical_names_cyclic_dependency_groups(
    temporary_directory: Path,
) -> None:
    """
    group-1 -> group-2
    group-2 -> group-1
    group-3 -> group-2
    """
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.GROUP_1]
include-groups = [
    "gRoup_2",
]

[tool.poetry.group.GROUP_1.dependencies]
foo = "*"

[tool.poetry.group.group_2]
include-groups = [
    "groUp_1",
]
[tool.poetry.group.group_2.dependencies]
bar = "*"

[tool.poetry.group.group_3]
include-groups = [
    "group_2",
]
[tool.poetry.group.group_3.dependencies]
baz = "*"
"""

    expected = """\
The Poetry configuration is invalid:
  - Cyclic dependency group include in group-1: group-2 -> group-1
  - Cyclic dependency group include in group-2: group-1 -> group-2
  - Cyclic dependency group include in group-3: group-2 -> group-1 -> group-2
"""
    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=RuntimeError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_unknown_nested_dependency_groups(
    temporary_directory: Path,
) -> None:
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.dev]
include-groups = [
    "testing",
]
[tool.poetry.group.dev.dependencies]
black = "*"
"""
    expected = "Group 'dev' includes group 'testing' which is not defined."

    assert_invalid_group_including(
        toml_data=content,
        expected_error=expected,
        error_type=ValueError,
        temporary_directory=temporary_directory,
    )


def test_create_poetry_with_included_groups_only(temporary_directory: Path) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.lint.dependencies]
black = "*"

[tool.poetry.group.testing.dependencies]
pytest = "*"

[tool.poetry.group.all]
include-groups = [
    "lint",
    "testing",
]
"""
    pyproject_toml.write_text(content, encoding="utf-8")

    poetry = Factory().create_poetry(temporary_directory)
    assert len(poetry.package.all_requires) == 4
    assert [
        (dep.name, ",".join(dep.groups)) for dep in poetry.package.all_requires
    ] == [
        ("black", "lint"),
        ("pytest", "testing"),
        ("black", "all"),
        ("pytest", "all"),
    ]


def test_create_poetry_with_nested_similar_dependencies(
    temporary_directory: Path,
) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"
    content = """\
[project]
name = "my-package"
version = "1.2.3"

[tool.poetry.group.parent.dependencies]
foo = "*"

[tool.poetry.group.parent]
include-groups = [
    "child",
]

[tool.poetry.group.child.dependencies]
foo = "*"

"""

    pyproject_toml.write_text(content, encoding="utf-8")

    poetry = Factory().create_poetry(temporary_directory)
    assert len(poetry.package.all_requires) == 3
    assert [
        (dep.name, ",".join(dep.groups)) for dep in poetry.package.all_requires
    ] == [
        # Duplicates because dependency is included via several groups.
        # This is ok because they are merged during dependency resolution.
        ("foo", "parent"),
        ("foo", "parent"),
        ("foo", "child"),
    ]


@pytest.mark.parametrize("optional", [False, True])
def test_create_poetry_with_optional_dependency_group(
    temporary_directory: Path, optional: bool
) -> None:
    pyproject_toml = temporary_directory / "pyproject.toml"
    content = f"""\
[project]
name = "my-package"
version = "1.2.3"

[dependency-groups]
dev = [ "foo" ]

# only "optional" in poetry section, no dependencies or include-groups
[tool.poetry.group.dev]
optional = {str(optional).lower()}
"""

    pyproject_toml.write_text(content, encoding="utf-8")

    poetry = Factory().create_poetry(temporary_directory)
    assert len(poetry.package.all_requires) == 1
    assert poetry.package.has_dependency_group("dev")
    assert poetry.package.dependency_group("dev").is_optional() is optional



from __future__ import annotations

import shutil
import subprocess
import tarfile
import tempfile
import zipfile

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import tomli_w

from poetry.core.utils._compat import tomllib


if TYPE_CHECKING:
    from collections.abc import Generator


__toml_build_backend_patch__ = {
    "build-system": {
        "requires": [str(Path(__file__).parent.parent)],
        "build-backend": "poetry.core.masonry.api",
    }
}


@contextmanager
def temporary_project_directory(
    path: Path, toml_patch: dict[str, Any] | None = None
) -> Generator[str, None, None]:
    """
    Context manager that takes a project source directory, copies content to a temporary
    directory, patches the `pyproject.toml` using the provided patch, or using the
    default patch if one is not given. The default path replaces `build-system` section
    in order to use the working copy of poetry-core as the backend.

    Once the context, exists, the temporary directory is cleaned up.

    :param path: Source project root directory to copy from.
    :param toml_patch: Patch to use for the pyproject.toml,
                        defaults to build system patching.
    :return: A temporary copy
    """
    assert (path / "pyproject.toml").exists()

    with tempfile.TemporaryDirectory(
        prefix="poetry-core-pep517", ignore_cleanup_errors=True
    ) as tmp:
        dst = Path(tmp) / path.name
        shutil.copytree(str(path), dst)
        toml = dst / "pyproject.toml"
        with toml.open("rb") as f:
            data = tomllib.load(f)
        data.update(toml_patch or __toml_build_backend_patch__)
        with toml.open("wb") as f:
            tomli_w.dump(data, f)
        yield str(dst)


def subprocess_run(*args: str, **kwargs: Any) -> subprocess.CompletedProcess[str]:
    """
    Helper method to run a subprocess. Asserts for success.
    """
    result = subprocess.run(
        args, text=True, encoding="locale", capture_output=True, **kwargs
    )
    assert result.returncode == 0
    return result


def validate_wheel_contents(
    name: str, version: str, path: Path, files: list[str] | None = None
) -> None:
    dist_info = f"{name}-{version}.dist-info"
    files = files or []

    with zipfile.ZipFile(path) as z:
        namelist = z.namelist()
        for filename in ["WHEEL", "METADATA", "RECORD", *files]:
            assert f"{dist_info}/{filename}" in namelist


def validate_sdist_contents(
    name: str, version: str, path: Path, files: list[str]
) -> None:
    escaped_name = name.replace("-", "_")
    with tarfile.open(path) as tar:
        namelist = tar.getnames()
        for filename in files:
            assert f"{escaped_name}-{version}/{filename}" in namelist



from __future__ import annotations

import pytest

from poetry.core.packages.specification import PackageSpecification


@pytest.mark.parametrize(
    "spec1, spec2, expected",
    [
        (PackageSpecification("a"), PackageSpecification("a"), True),
        (PackageSpecification("a", "type1"), PackageSpecification("a", "type1"), True),
        (PackageSpecification("a", "type1"), PackageSpecification("a", "type2"), False),
        (PackageSpecification("a"), PackageSpecification("a", "type1"), False),
        (PackageSpecification("a", "type1"), PackageSpecification("a"), False),
    ],
)
def test_is_same_package_source_type(
    spec1: PackageSpecification,
    spec2: PackageSpecification,
    expected: bool,
) -> None:
    assert spec1.is_same_package_as(spec2) == expected


@pytest.mark.parametrize(
    ("source_type", "result"),
    [
        ("directory", True),
        ("file", True),
        ("url", True),
        ("git", True),
        ("legacy", False),
        (None, False),
    ],
)
def test_is_direct_origin(source_type: str | None, result: bool) -> None:
    assert PackageSpecification("package", source_type).is_direct_origin() == result


@pytest.mark.parametrize(
    "spec1, spec2, expected",
    [
        (PackageSpecification("a"), PackageSpecification("a"), True),
        (PackageSpecification("a"), PackageSpecification("b"), False),
        (PackageSpecification("a", features=["x"]), PackageSpecification("a"), True),
        (
            PackageSpecification("a", features=["x"]),
            PackageSpecification("a", features=["x"]),
            True,
        ),
        (
            PackageSpecification("a", features=["x"]),
            PackageSpecification("b", features=["x"]),
            False,
        ),
        (
            PackageSpecification("a", features=["x"]),
            PackageSpecification("a", features=["y"]),
            False,
        ),
        (
            PackageSpecification("a", features=["x"]),
            PackageSpecification("a", features=["x", "y"]),
            False,
        ),
        (
            PackageSpecification("a", features=["x", "y"]),
            PackageSpecification("a", features=["x"]),
            True,
        ),
    ],
)
def test_specification_provides(
    spec1: PackageSpecification,
    spec2: PackageSpecification,
    expected: bool,
) -> None:
    assert spec1.provides(spec2) == expected


@pytest.mark.parametrize(
    "spec1, spec2",
    [
        (
            # nothing except for name and features matters if no source
            PackageSpecification("a", None, "url1", "ref1", "resref1", "sub1"),
            PackageSpecification("a", None, "url2", "ref2", "resref2", "sub2"),
        ),
        (
            # ref does not matter if resolved ref is equal
            PackageSpecification("a", "type", "url", "ref1", "resref1"),
            PackageSpecification("a", "type", "url", "ref2", "resref1"),
        ),
        (
            # resolved ref does not matter if no ref
            PackageSpecification("a", "type", "url", None, "resref1"),
            PackageSpecification("a", "type", "url", None, "resref2"),
        ),
        (
            # resolved ref unset when ref starts with other
            PackageSpecification("a", "type", "url", "ref/a", "resref1"),
            PackageSpecification("a", "type", "url", "ref", None),
        ),
        (
            # resolved ref unset when ref starts with other
            PackageSpecification("a", "type", "url", "ref/a", None),
            PackageSpecification("a", "type", "url", "ref", "resref2"),
        ),
    ],
)
def test_equal_specifications_have_same_hash(
    spec1: PackageSpecification, spec2: PackageSpecification
) -> None:
    assert spec1 == spec2
    assert spec2 == spec1
    assert hash(spec1) == hash(spec2)


@pytest.mark.parametrize(
    "source_url,normalized_url",
    [
        ("https://github.com/demo/demo.git", "https://github.com/demo/demo.git"),
        ("git@github.com:demo/demo.git", "ssh://git@github.com/demo/demo.git"),
    ],
)
def test_specification_normalize_source_url_method(
    source_url: str, normalized_url: str
) -> None:
    assert (
        PackageSpecification._normalize_source_url("git", source_url) == normalized_url
    )
    assert (
        PackageSpecification._normalize_source_url("notgit", source_url) == source_url
    )


@pytest.mark.parametrize(
    "source_url,normalized_url",
    [
        ("https://github.com/demo/demo.git", "https://github.com/demo/demo.git"),
        ("git@github.com:demo/demo.git", "ssh://git@github.com/demo/demo.git"),
    ],
)
def test_specification_uses_normalize_source_url_for_git(
    source_url: str, normalized_url: str
) -> None:
    assert (
        PackageSpecification(
            name="demo",
            source_type="git",
            source_url=source_url,
        ).source_url
        == normalized_url
    )



from __future__ import annotations

import sys

from pathlib import Path
from typing import TYPE_CHECKING
from typing import cast

import pytest

from poetry.core.packages.dependency import Dependency
from poetry.core.packages.file_dependency import FileDependency
from poetry.core.version.markers import SingleMarker


if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture

    from poetry.core.version.markers import BaseMarker

DIST_PATH = Path(__file__).parent.parent / "fixtures" / "distributions"
TEST_FILE = "demo-0.1.0.tar.gz"


def test_file_dependency_does_not_exist(
    caplog: LogCaptureFixture, mocker: MockerFixture
) -> None:
    mock_exists = mocker.patch.object(Path, "exists")
    mock_exists.return_value = False
    dep = FileDependency("demo", DIST_PATH / "demo-0.2.0.tar.gz")
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "does not exist" in record.message

    with pytest.raises(ValueError, match="does not exist"):
        dep.validate(raise_error=True)

    mock_exists.assert_called_once()


def test_file_dependency_is_directory(
    caplog: LogCaptureFixture, mocker: MockerFixture
) -> None:
    mock_is_directory = mocker.patch.object(Path, "is_dir")
    mock_is_directory.return_value = True
    dep = FileDependency("demo", DIST_PATH)
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "is a directory" in record.message

    with pytest.raises(ValueError, match="is a directory"):
        dep.validate(raise_error=True)

    mock_is_directory.assert_called_once()


def _test_file_dependency_pep_508(
    mocker: MockerFixture,
    name: str,
    path: Path,
    pep_508_input: str,
    pep_508_output: str | None = None,
    marker: BaseMarker | None = None,
) -> None:
    mocker.patch.object(Path, "exists").return_value = True
    mocker.patch.object(Path, "is_file").return_value = True

    dep = Dependency.create_from_pep_508(
        pep_508_input, relative_to=Path(__file__).parent
    )
    if marker:
        dep.marker = marker

    assert dep.is_file()
    dep = cast("FileDependency", dep)
    assert dep.name == name
    assert dep.path == path
    assert dep.to_pep_508() == (pep_508_output or pep_508_input)


def test_file_dependency_pep_508_local_file_absolute(mocker: MockerFixture) -> None:
    path = DIST_PATH / "demo-0.2.0.tar.gz"
    expected = f"demo @ {path.as_uri()}"

    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo @ file://{prefix}{path.as_posix()}"
    _test_file_dependency_pep_508(mocker, "demo", path, requirement, expected)

    requirement = f"demo @ {path}"
    _test_file_dependency_pep_508(mocker, "demo", path, requirement, expected)


def test_file_dependency_pep_508_local_file_localhost(mocker: MockerFixture) -> None:
    path = DIST_PATH / "demo-0.2.0.tar.gz"
    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo @ file://localhost{prefix}{path.as_posix()}"
    expected = f"demo @ {path.as_uri()}"
    _test_file_dependency_pep_508(mocker, "demo", path, requirement, expected)


def test_file_dependency_pep_508_local_file_relative_path(
    mocker: MockerFixture,
) -> None:
    path = Path("..") / "fixtures" / "distributions" / "demo-0.2.0.tar.gz"

    with pytest.raises(ValueError):
        requirement = f"demo @ file://{path.as_posix()}"
        _test_file_dependency_pep_508(mocker, "demo", path, requirement)

    requirement = f"demo @ {path}"
    base = Path(__file__).parent
    expected = f"demo @ {(base / path).resolve().as_uri()}"
    _test_file_dependency_pep_508(mocker, "demo", path, requirement, expected)


def test_file_dependency_pep_508_with_subdirectory(mocker: MockerFixture) -> None:
    path = DIST_PATH / "demo.zip"
    expected = f"demo @ {path.as_uri()}#subdirectory=sub"

    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo @ file://{prefix}{path.as_posix()}#subdirectory=sub"
    _test_file_dependency_pep_508(mocker, "demo", path, requirement, expected)


def test_to_pep_508_with_marker(mocker: MockerFixture) -> None:
    wheel = "demo-0.1.0-py2.py3-none-any.whl"

    abs_path = DIST_PATH / wheel
    requirement = f'demo @ {abs_path.as_uri()} ; sys_platform == "linux"'
    _test_file_dependency_pep_508(
        mocker,
        "demo",
        abs_path,
        requirement,
        marker=SingleMarker("sys.platform", "linux"),
    )


def test_relative_file_dependency_to_pep_508_with_marker(mocker: MockerFixture) -> None:
    wheel = "demo-0.1.0-py2.py3-none-any.whl"

    rel_path = Path("..") / "fixtures" / "distributions" / wheel
    requirement = f'demo @ {rel_path.as_posix()} ; sys_platform == "linux"'
    base = Path(__file__).parent
    expected = (
        f'demo @ {(base / rel_path).resolve().as_uri()} ; sys_platform == "linux"'
    )
    _test_file_dependency_pep_508(
        mocker,
        "demo",
        rel_path,
        requirement,
        expected,
        marker=SingleMarker("sys.platform", "linux"),
    )


def test_file_dependency_pep_508_extras(mocker: MockerFixture) -> None:
    wheel = "demo-0.1.0-py2.py3-none-any.whl"

    rel_path = Path("..") / "fixtures" / "distributions" / wheel
    requirement = f'demo[foo,bar] @ {rel_path.as_posix()} ; sys_platform == "linux"'
    base = Path(__file__).parent
    expected = (
        f"demo[bar,foo] @ {(base / rel_path).resolve().as_uri()} ;"
        ' sys_platform == "linux"'
    )
    _test_file_dependency_pep_508(
        mocker,
        "demo",
        rel_path,
        requirement,
        expected,
    )


@pytest.mark.parametrize(
    "name,path,extras,constraint,expected",
    [
        (
            "demo",
            DIST_PATH / TEST_FILE,
            None,
            None,
            f"demo (*) @ {(DIST_PATH / TEST_FILE).as_uri()}",
        ),
        (
            "demo",
            DIST_PATH / TEST_FILE,
            ["foo"],
            "1.2",
            f"demo[foo] (1.2) @ {(DIST_PATH / TEST_FILE).as_uri()}",
        ),
    ],
)
def test_file_dependency_string_representation(
    name: str,
    path: Path,
    extras: list[str] | None,
    constraint: str | None,
    expected: str,
) -> None:
    dependency = FileDependency(name=name, path=path, extras=extras)
    if constraint:
        dependency.constraint = constraint
    assert str(dependency) == expected



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import Sequence

    from poetry.core.masonry.utils.include import Include


class ModuleOrPackageNotFoundError(ValueError):
    pass


class Module:
    def __init__(
        self,
        name: str,
        directory: str = ".",
        packages: Sequence[Mapping[str, Any]] = (),
        includes: Sequence[Mapping[str, Any]] = (),
    ) -> None:
        from poetry.core.masonry.utils.include import Include
        from poetry.core.masonry.utils.package_include import PackageInclude
        from poetry.core.utils.helpers import module_name

        self._name = module_name(name)
        self._in_src = False
        self._is_package = False
        self._path = Path(directory)
        self._package_includes: list[PackageInclude] = []
        self._explicit_includes: list[Include] = []

        if not packages:
            # It must exist either as a .py file or a directory, but not both
            pkg_dir = Path(directory, self._name)
            py_file = Path(directory, self._name + ".py")
            default_package: dict[str, Any]
            if pkg_dir.is_dir() and py_file.is_file():
                raise ValueError(f"Both {pkg_dir} and {py_file} exist")
            elif pkg_dir.is_dir():
                default_package = {"include": str(pkg_dir.relative_to(self._path))}
            elif py_file.is_file():
                default_package = {"include": str(py_file.relative_to(self._path))}
            else:
                # Searching for a src module
                src = Path(directory, "src")
                src_pkg_dir = src / self._name
                src_py_file = src / (self._name + ".py")

                if src_pkg_dir.is_dir() and src_py_file.is_file():
                    raise ValueError(f"Both {pkg_dir} and {py_file} exist")
                elif src_pkg_dir.is_dir():
                    default_package = {
                        "include": str(src_pkg_dir.relative_to(src)),
                        "from": str(src.relative_to(self._path)),
                    }
                elif src_py_file.is_file():
                    default_package = {
                        "include": str(src_py_file.relative_to(src)),
                        "from": str(src.relative_to(self._path)),
                    }
                else:
                    raise ModuleOrPackageNotFoundError(
                        f"No file/folder found for package {name}"
                    )
            default_package["format"] = ["sdist", "wheel"]
            packages = [default_package]

        for package in packages:
            self._package_includes.append(
                PackageInclude(
                    self._path,
                    package["include"],
                    formats=package["format"],
                    source=package.get("from"),
                    target=package.get("to"),
                )
            )

        for include in includes:
            self._explicit_includes.append(
                Include(self._path, include["path"], formats=include["format"])
            )

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> Path:
        return self._path

    @property
    def file(self) -> Path:
        if self._is_package:
            return self._path / "__init__.py"
        else:
            return self._path

    @property
    def includes(self) -> list[Include]:
        return [*self._package_includes, *self._explicit_includes]

    @property
    def explicit_includes(self) -> list[Include]:
        return self._explicit_includes

    def is_package(self) -> bool:
        return self._is_package

    def is_in_src(self) -> bool:
        return self._in_src



"""
PEP-517 compliant buildsystem API
"""

from __future__ import annotations

import logging

from pathlib import Path
from typing import Any

from poetry.core.factory import Factory
from poetry.core.masonry.builders.sdist import SdistBuilder
from poetry.core.masonry.builders.wheel import WheelBuilder


log = logging.getLogger(__name__)


def get_requires_for_build_wheel(
    config_settings: dict[str, Any] | None = None,
) -> list[str]:
    """
    Returns an additional list of requirements for building, as PEP508 strings,
    above and beyond those specified in the pyproject.toml file.

    This implementation is optional. At the moment it only returns an empty list,
    which would be the same as if not define. So this is just for completeness
    for future implementation.
    """

    return []


# For now, we require all dependencies to build either a wheel or an sdist.
get_requires_for_build_sdist = get_requires_for_build_wheel


def prepare_metadata_for_build_wheel(
    metadata_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    poetry = Factory().create_poetry(Path().resolve(), with_groups=False)
    builder = WheelBuilder(poetry, config_settings=config_settings)
    metadata_path = Path(metadata_directory)
    dist_info = builder.prepare_metadata(metadata_path)
    return dist_info.name


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    """Builds a wheel, places it in wheel_directory"""
    poetry = Factory().create_poetry(Path().resolve(), with_groups=False)
    metadata_path = None if metadata_directory is None else Path(metadata_directory)

    return WheelBuilder.make_in(
        poetry,
        Path(wheel_directory),
        metadata_directory=metadata_path,
        config_settings=config_settings,
    )


def build_sdist(
    sdist_directory: str, config_settings: dict[str, Any] | None = None
) -> str:
    """Builds an sdist, places it in sdist_directory"""
    poetry = Factory().create_poetry(Path().resolve(), with_groups=False)

    path = SdistBuilder(poetry, config_settings=config_settings).build(
        Path(sdist_directory)
    )

    return path.name


def build_editable(
    wheel_directory: str,
    config_settings: dict[str, Any] | None = None,
    metadata_directory: str | None = None,
) -> str:
    poetry = Factory().create_poetry(Path().resolve(), with_groups=False)
    metadata_path = None if metadata_directory is None else Path(metadata_directory)

    return WheelBuilder.make_in(
        poetry,
        Path(wheel_directory),
        metadata_directory=metadata_path,
        editable=True,
        config_settings=config_settings,
    )


get_requires_for_build_editable = get_requires_for_build_wheel
prepare_metadata_for_build_editable = prepare_metadata_for_build_wheel



from __future__ import annotations


class BaseConstraint:
    def allows(self, other: BaseConstraint) -> bool:
        raise NotImplementedError

    def allows_all(self, other: BaseConstraint) -> bool:
        raise NotImplementedError

    def allows_any(self, other: BaseConstraint) -> bool:
        raise NotImplementedError

    def invert(self) -> BaseConstraint:
        raise NotImplementedError()

    def difference(self, other: BaseConstraint) -> BaseConstraint:
        raise NotImplementedError

    def intersect(self, other: BaseConstraint) -> BaseConstraint:
        raise NotImplementedError

    def union(self, other: BaseConstraint) -> BaseConstraint:
        raise NotImplementedError

    def is_any(self) -> bool:
        return False

    def is_empty(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self}>"

    def __str__(self) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError



from __future__ import annotations

import subprocess

from typing import TYPE_CHECKING

from poetry.core.vcs.git import Git


if TYPE_CHECKING:
    from pathlib import Path


def get_vcs(directory: Path) -> Git | None:
    directory = directory.resolve(strict=True)
    vcs: Git | None

    try:
        from poetry.core.vcs.git import executable

        check_ignore = subprocess.run(
            [executable(), "check-ignore", "."],
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            cwd=directory,
        ).returncode

        if check_ignore == 0:
            vcs = None
        else:
            rel_path_to_git_dir = subprocess.check_output(
                [executable(), "rev-parse", "--show-cdup"],
                stderr=subprocess.STDOUT,
                text=True,
                encoding="utf-8",
                cwd=directory,
            ).strip()

            vcs = Git((directory / rel_path_to_git_dir).resolve())

    except (subprocess.CalledProcessError, OSError, RuntimeError):
        vcs = None

    return vcs



from __future__ import annotations

import pytest

from poetry.core.constraints.version import Version
from poetry.core.constraints.version import VersionConstraint
from poetry.core.constraints.version import VersionRange
from poetry.core.constraints.version import VersionUnion
from poetry.core.constraints.version import parse_constraint
from poetry.core.constraints.version import parse_marker_version_constraint
from poetry.core.constraints.version.exceptions import ParseConstraintError
from poetry.core.version.pep440 import ReleaseTag


@pytest.mark.parametrize(
    "input,constraint",
    [
        ("*", VersionRange()),
        ("*.*", VersionRange()),
        ("v*.*", VersionRange()),
        (">1.0.0", VersionRange(min=Version.from_parts(1, 0, 0))),
        ("<1.2.3", VersionRange(max=Version.from_parts(1, 2, 3))),
        ("<=1.2.3", VersionRange(max=Version.from_parts(1, 2, 3), include_max=True)),
        (">=1.2.3", VersionRange(min=Version.from_parts(1, 2, 3), include_min=True)),
        ("=1.2.3", Version.from_parts(1, 2, 3)),
        ("1.2.3", Version.from_parts(1, 2, 3)),
        ("1!2.3.4", Version.from_parts(2, 3, 4, epoch=1)),
        ("=1.0", Version.from_parts(1, 0, 0)),
        ("1.2.3b5", Version.from_parts(1, 2, 3, pre=ReleaseTag("beta", 5))),
        (
            "1.0.0a1.dev0",
            Version.from_parts(
                1, 0, 0, pre=ReleaseTag("a", 1), dev=ReleaseTag("dev", 0)
            ),
        ),
        (
            ">dev",
            VersionRange(min=Version.from_parts(0, 0, dev=ReleaseTag("dev"))),
        ),  # Issue 206
    ],
)
def test_parse_constraint(input: str, constraint: Version | VersionRange) -> None:
    assert parse_constraint(input) == constraint


@pytest.mark.parametrize(
    "input,constraint",
    [
        (
            "v2.*",
            VersionRange(Version.parse("2.dev0"), Version.parse("3.dev0"), True),
        ),
        (
            "2.*.*",
            VersionRange(Version.parse("2.dev0"), Version.parse("3.dev0"), True),
        ),
        (
            "20.*",
            VersionRange(Version.parse("20.dev0"), Version.parse("21.dev0"), True),
        ),
        (
            "20.*.*",
            VersionRange(Version.parse("20.dev0"), Version.parse("21.dev0"), True),
        ),
        (
            "2.0.*",
            VersionRange(Version.parse("2.0.dev0"), Version.parse("2.1.dev0"), True),
        ),
        ("0.*", VersionRange(Version.parse("0.dev0"), Version.parse("1.dev0"), True)),
        ("0.*.*", VersionRange(Version.parse("0.dev0"), Version.parse("1.dev0"), True)),
        (
            "2.0.post1.*",
            VersionRange(
                min=Version.parse("2.0.post1.dev0"),
                max=Version.parse("2.0.post2.dev0"),
                include_min=True,
                include_max=False,
            ),
        ),
        (
            "2.0a1.*",
            VersionRange(
                min=Version.parse("2.0a1.dev0"),
                max=Version.parse("2.0a2.dev0"),
                include_min=True,
                include_max=False,
            ),
        ),
        (
            "2.0dev0.*",
            VersionRange(
                min=Version.parse("2.0dev0"),
                max=Version.parse("2.0dev1"),
                include_min=True,
                include_max=False,
            ),
        ),
    ],
)
def test_parse_constraint_wildcard(input: str, constraint: VersionRange) -> None:
    assert parse_constraint(input) == constraint


@pytest.mark.parametrize(
    "input,constraint",
    [
        (
            "~v1",
            VersionRange(
                Version.from_parts(1, 0, 0), Version.from_parts(2, 0, 0), True
            ),
        ),
        (
            "~1.0",
            VersionRange(
                Version.from_parts(1, 0, 0), Version.from_parts(1, 1, 0), True
            ),
        ),
        (
            "~1.0.0",
            VersionRange(
                Version.from_parts(1, 0, 0), Version.from_parts(1, 1, 0), True
            ),
        ),
        (
            "~1.2",
            VersionRange(
                Version.from_parts(1, 2, 0), Version.from_parts(1, 3, 0), True
            ),
        ),
        (
            "~1.2.3",
            VersionRange(
                Version.from_parts(1, 2, 3), Version.from_parts(1, 3, 0), True
            ),
        ),
        (
            "~1.0.0a1",
            VersionRange(
                min=Version.from_parts(1, 0, 0, pre=ReleaseTag("a", 1)),
                max=Version.from_parts(1, 1, 0),
                include_min=True,
            ),
        ),
        (
            "~1.0.0a1.dev0",
            VersionRange(
                min=Version.from_parts(
                    1, 0, 0, pre=ReleaseTag("a", 1), dev=ReleaseTag("dev", 0)
                ),
                max=Version.from_parts(1, 1, 0),
                include_min=True,
            ),
        ),
        (
            "~1.2-beta",
            VersionRange(
                Version.from_parts(1, 2, 0, pre=ReleaseTag("beta")),
                Version.from_parts(1, 3, 0),
                True,
            ),
        ),
        (
            "~1.2-b2",
            VersionRange(
                Version.from_parts(1, 2, 0, pre=ReleaseTag("beta", 2)),
                Version.from_parts(1, 3, 0),
                True,
            ),
        ),
        (
            "~0.3",
            VersionRange(
                Version.from_parts(0, 3, 0), Version.from_parts(0, 4, 0), True
            ),
        ),
        (
            "~3.5",
            VersionRange(
                Version.from_parts(3, 5, 0), Version.from_parts(3, 6, 0), True
            ),
        ),
        (
            "~=3.5",
            VersionRange(
                Version.from_parts(3, 5, 0), Version.from_parts(4, 0, 0), True
            ),
        ),  # PEP 440
        (
            "~=3.5.3",
            VersionRange(
                Version.from_parts(3, 5, 3), Version.from_parts(3, 6, 0), True
            ),
        ),  # PEP 440
        (
            "~=3.5.3rc1",
            VersionRange(
                Version.from_parts(3, 5, 3, pre=ReleaseTag("rc", 1)),
                Version.from_parts(3, 6, 0),
                True,
            ),
        ),  # PEP 440
    ],
)
def test_parse_constraint_tilde(input: str, constraint: VersionRange) -> None:
    assert parse_constraint(input) == constraint


@pytest.mark.parametrize(
    "input,constraint",
    [
        (
            "^v1",
            VersionRange(
                Version.from_parts(1, 0, 0), Version.from_parts(2, 0, 0), True
            ),
        ),
        ("^0", VersionRange(Version.from_parts(0), Version.from_parts(1), True)),
        (
            "^0.0",
            VersionRange(
                Version.from_parts(0, 0, 0), Version.from_parts(0, 1, 0), True
            ),
        ),
        (
            "^1.2",
            VersionRange(
                Version.from_parts(1, 2, 0), Version.from_parts(2, 0, 0), True
            ),
        ),
        (
            "^1.2.3-beta.2",
            VersionRange(
                Version.from_parts(1, 2, 3, pre=ReleaseTag("beta", 2)),
                Version.from_parts(2, 0, 0),
                True,
            ),
        ),
        (
            "^1.2.3",
            VersionRange(
                Version.from_parts(1, 2, 3), Version.from_parts(2, 0, 0), True
            ),
        ),
        (
            "^0.2.3",
            VersionRange(
                Version.from_parts(0, 2, 3), Version.from_parts(0, 3, 0), True
            ),
        ),
        (
            "^0.2",
            VersionRange(
                Version.from_parts(0, 2, 0), Version.from_parts(0, 3, 0), True
            ),
        ),
        (
            "^0.2.0",
            VersionRange(
                Version.from_parts(0, 2, 0), Version.from_parts(0, 3, 0), True
            ),
        ),
        (
            "^0.0.3",
            VersionRange(
                Version.from_parts(0, 0, 3), Version.from_parts(0, 0, 4), True
            ),
        ),
        (
            "^0.0.3-alpha.21",
            VersionRange(
                Version.from_parts(0, 0, 3, pre=ReleaseTag("alpha", 21)),
                Version.from_parts(0, 0, 4),
                True,
            ),
        ),
        (
            "^0.1.3-alpha.21",
            VersionRange(
                Version.from_parts(0, 1, 3, pre=ReleaseTag("alpha", 21)),
                Version.from_parts(0, 2, 0),
                True,
            ),
        ),
        (
            "^0.0.0-alpha.21",
            VersionRange(
                Version.from_parts(0, 0, 0, pre=ReleaseTag("alpha", 21)),
                Version.from_parts(0, 0, 1),
                True,
            ),
        ),
        (
            "^1.0.0a1",
            VersionRange(
                min=Version.from_parts(1, 0, 0, pre=ReleaseTag("a", 1)),
                max=Version.from_parts(2, 0, 0),
                include_min=True,
            ),
        ),
        (
            "^1.0.0a1.dev0",
            VersionRange(
                min=Version.from_parts(
                    1, 0, 0, pre=ReleaseTag("a", 1), dev=ReleaseTag("dev", 0)
                ),
                max=Version.from_parts(2, 0, 0),
                include_min=True,
            ),
        ),
    ],
)
def test_parse_constraint_caret(input: str, constraint: VersionRange) -> None:
    assert parse_constraint(input) == constraint


def test_parse_constraint_multi() -> None:
    assert parse_constraint(">2.0,<=3.0") == VersionRange(
        Version.from_parts(2, 0, 0),
        Version.from_parts(3, 0, 0),
        include_min=False,
        include_max=True,
    )


@pytest.mark.parametrize(
    "input, output",
    [
        (
            ">1!2,<=2!3",
            VersionRange(
                Version.from_parts(2, 0, 0, epoch=1),
                Version.from_parts(3, 0, 0, epoch=2),
                include_min=False,
                include_max=True,
            ),
        ),
        (
            ">=1!2,<2!3",
            VersionRange(
                Version.from_parts(2, 0, 0, epoch=1),
                Version.from_parts(3, 0, 0, epoch=2),
                include_min=True,
                include_max=False,
            ),
        ),
    ],
)
def test_parse_constraint_multi_with_epochs(input: str, output: VersionRange) -> None:
    assert parse_constraint(input) == output


def test_parse_constraint_multi_wildcard() -> None:
    assert parse_constraint(">=2.7,!=3.0.*,!=3.1.*") == VersionUnion(
        VersionRange(Version.parse("2.7"), Version.parse("3.0.dev0"), True, False),
        VersionRange(Version.parse("3.2.dev0"), None, True, False),
    )


@pytest.mark.parametrize(
    "input,constraint",
    [
        (
            "!=v2.*",
            VersionRange(max=Version.parse("2.0.0.dev0")).union(
                VersionRange(Version.parse("3.0.dev0"), include_min=True)
            ),
        ),
        (
            "!=2.*.*",
            VersionRange(max=Version.parse("2.0.0.dev0")).union(
                VersionRange(Version.parse("3.0.dev0"), include_min=True)
            ),
        ),
        (
            "!=2.0.*",
            VersionRange(max=Version.parse("2.0.0.dev0")).union(
                VersionRange(Version.parse("2.1.dev0"), include_min=True)
            ),
        ),
        (
            "!=0.*",
            VersionRange(max=Version.parse("0.dev0")).union(
                VersionRange(Version.parse("1.0.dev0"), include_min=True)
            ),
        ),
        (
            "!=0.*.*",
            VersionRange(max=Version.parse("0.dev0")).union(
                VersionRange(Version.parse("1.0.dev0"), include_min=True)
            ),
        ),
    ],
)
def test_parse_constraints_negative_wildcard(
    input: str, constraint: VersionRange
) -> None:
    assert parse_constraint(input) == constraint


@pytest.mark.parametrize(
    "input,constraint",
    [
        (">3.7,", VersionRange(min=Version.parse("3.7"))),
        (">3.7 , ", VersionRange(min=Version.parse("3.7"))),
        (
            ">3.7,<3.8,",
            VersionRange(min=Version.parse("3.7"), max=Version.parse("3.8")),
        ),
        (
            ">3.7,||<3.6,",
            VersionRange(min=Version.parse("3.7")).union(
                VersionRange(max=Version.parse("3.6"))
            ),
        ),
        (
            ">3.7 , || <3.6 , ",
            VersionRange(min=Version.parse("3.7")).union(
                VersionRange(max=Version.parse("3.6"))
            ),
        ),
        (
            ">3.7, <3.8, || <3.6, >3.5",
            VersionRange(min=Version.parse("3.7"), max=Version.parse("3.8")).union(
                VersionRange(min=Version.parse("3.5"), max=Version.parse("3.6"))
            ),
        ),
    ],
)
def test_parse_constraints_with_trailing_comma(
    input: str, constraint: VersionRange
) -> None:
    assert parse_constraint(input) == constraint


@pytest.mark.parametrize(
    "input, expected",
    [
        ("1", "1"),
        ("1.2", "1.2"),
        ("1.2.3", "1.2.3"),
        ("!=1", "!=1"),
        ("!=1.2", "!=1.2"),
        ("!=1.2.3", "!=1.2.3"),
        ("^1", ">=1,<2"),
        ("^1.0", ">=1.0,<2.0"),
        ("^1.0.0", ">=1.0.0,<2.0.0"),
        ("^1.0.0-alpha.1", ">=1.0.0a1,<2.0.0"),
        ("^0", ">=0,<1"),
        ("^0.1", ">=0.1,<0.2"),
        ("^0.0.2", ">=0.0.2,<0.0.3"),
        ("^0.1.2", ">=0.1.2,<0.2.0"),
        ("^0-alpha.1", ">=0a1,<1"),
        ("^0.1-alpha.1", ">=0.1a1,<0.2"),
        ("^0.0.2-alpha.1", ">=0.0.2a1,<0.0.3"),
        ("^0.1.2-alpha.1", ">=0.1.2a1,<0.2.0"),
        ("~1", ">=1,<2"),
        ("~1.0", ">=1.0,<1.1"),
        ("~1.0.0", ">=1.0.0,<1.1.0"),
    ],
)
def test_constraints_keep_version_precision(input: str, expected: str) -> None:
    assert str(parse_constraint(input)) == expected


@pytest.mark.parametrize(
    "constraint_parts,expected",
    [
        (["3.8"], Version.from_parts(3, 8)),
        (["=", "3.8"], Version.from_parts(3, 8)),
        (["==", "3.8"], Version.from_parts(3, 8)),
        ([">", "3.8"], VersionRange(min=Version.from_parts(3, 8))),
        ([">=", "3.8"], VersionRange(min=Version.from_parts(3, 8), include_min=True)),
        (["<", "3.8"], VersionRange(max=Version.from_parts(3, 8))),
        (["<=", "3.8"], VersionRange(max=Version.from_parts(3, 8), include_max=True)),
        (
            ["^", "3.8"],
            VersionRange(
                min=Version.from_parts(3, 8),
                max=Version.from_parts(4, 0),
                include_min=True,
            ),
        ),
        (
            ["~", "3.8"],
            VersionRange(
                min=Version.from_parts(3, 8),
                max=Version.from_parts(3, 9),
                include_min=True,
            ),
        ),
        (
            ["~=", "3.8"],
            VersionRange(
                min=Version.from_parts(3, 8),
                max=Version.from_parts(4, 0),
                include_min=True,
            ),
        ),
        (
            ["3.8.*"],
            VersionRange(
                min=Version.parse("3.8.0.dev0"),
                max=Version.parse("3.9.0.dev0"),
                include_min=True,
            ),
        ),
        (
            ["==", "3.8.*"],
            VersionRange(
                min=Version.parse("3.8.0.dev0"),
                max=Version.parse("3.9.0.dev0"),
                include_min=True,
            ),
        ),
        (
            ["!=", "3.8.*"],
            VersionRange(max=Version.parse("3.8.dev0")).union(
                VersionRange(Version.parse("3.9.dev0"), include_min=True)
            ),
        ),
        (
            [">", "3.8", ",", "<=", "6.5"],
            VersionRange(
                min=Version.from_parts(3, 8),
                max=Version.from_parts(6, 5),
                include_max=True,
            ),
        ),
        (
            [">=", "2.7", ",", "!=", "3.0.*", ",", "!=", "3.1.*"],
            VersionUnion(
                VersionRange(
                    Version.parse("2.7"), Version.parse("3.0.dev0"), True, False
                ),
                VersionRange(Version.parse("3.2.dev0"), None, True, False),
            ),
        ),
        (
            ["~", "2.7", "||", "~", "3.8"],
            VersionUnion(
                VersionRange(
                    min=Version.from_parts(2, 7),
                    max=Version.from_parts(2, 8),
                    include_min=True,
                ),
                VersionRange(
                    min=Version.from_parts(3, 8),
                    max=Version.from_parts(3, 9),
                    include_min=True,
                ),
            ),
        ),
        (
            ["~", "2.7", "||", "~", "3.8", "|", ">=", "3.10", ",", "<", "3.12"],
            VersionUnion(
                VersionRange(
                    min=Version.from_parts(2, 7),
                    max=Version.from_parts(2, 8),
                    include_min=True,
                ),
                VersionRange(
                    min=Version.from_parts(3, 8),
                    max=Version.from_parts(3, 9),
                    include_min=True,
                ),
                VersionRange(
                    min=Version.from_parts(3, 10),
                    max=Version.from_parts(3, 12),
                    include_min=True,
                ),
            ),
        ),
    ],
)
@pytest.mark.parametrize(("with_whitespace_padding",), [(True,), (False,)])
def test_parse_constraint_with_white_space_padding(
    constraint_parts: list[str],
    expected: VersionConstraint,
    with_whitespace_padding: bool,
) -> None:
    padding = " " * (4 if with_whitespace_padding else 0)
    constraint = padding.join(["", *constraint_parts, ""])
    assert parse_constraint(constraint) == expected


def test_parse_marker_constraint_does_not_allow_invalid_version() -> None:
    with pytest.raises(ParseConstraintError):
        parse_marker_version_constraint("4.9.253-tegra")


def test_parse_marker_constraint_does_allow_invalid_version_if_requested() -> None:
    assert parse_marker_version_constraint(
        "4.9.253-tegra", pep440=False
    ) == Version.from_parts(4, 9, 253, local="tegra")



"""Tree matcher based on Lark grammar"""

import re
from typing import List, Dict
from collections import defaultdict

from . import Tree, Token, Lark
from .common import ParserConf
from .exceptions import ConfigurationError
from .parsers import earley
from .grammar import Rule, Terminal, NonTerminal


def is_discarded_terminal(t):
    return t.is_term and t.filter_out


class _MakeTreeMatch:
    def __init__(self, name, expansion):
        self.name = name
        self.expansion = expansion

    def __call__(self, args):
        t = Tree(self.name, args)
        t.meta.match_tree = True
        t.meta.orig_expansion = self.expansion
        return t


def _best_from_group(seq, group_key, cmp_key):
    d = {}
    for item in seq:
        key = group_key(item)
        if key in d:
            v1 = cmp_key(item)
            v2 = cmp_key(d[key])
            if v2 > v1:
                d[key] = item
        else:
            d[key] = item
    return list(d.values())


def _best_rules_from_group(rules: List[Rule]) -> List[Rule]:
    rules = _best_from_group(rules, lambda r: r, lambda r: -len(r.expansion))
    rules.sort(key=lambda r: len(r.expansion))
    return rules


def _match(term, token):
    if isinstance(token, Tree):
        name, _args = parse_rulename(term.name)
        return token.data == name
    elif isinstance(token, Token):
        return term == Terminal(token.type)
    assert False, (term, token)


def make_recons_rule(origin, expansion, old_expansion):
    return Rule(origin, expansion, alias=_MakeTreeMatch(origin.name, old_expansion))


def make_recons_rule_to_term(origin, term):
    return make_recons_rule(origin, [Terminal(term.name)], [term])


def parse_rulename(s):
    "Parse rule names that may contain a template syntax (like rule{a, b, ...})"
    name, args_str = re.match(r'(\w+)(?:{(.+)})?', s).groups()
    args = args_str and [a.strip() for a in args_str.split(',')]
    return name, args



class ChildrenLexer:
    def __init__(self, children):
        self.children = children

    def lex(self, parser_state):
        return self.children

class TreeMatcher:
    """Match the elements of a tree node, based on an ontology
    provided by a Lark grammar.

    Supports templates and inlined rules (`rule{a, b,..}` and `_rule`)

    Initialize with an instance of Lark.
    """
    rules_for_root: Dict[str, List[Rule]]
    rules: List[Rule]
    parser: Lark

    def __init__(self, parser: Lark):
        # XXX TODO calling compile twice returns different results!
        assert not parser.options.maybe_placeholders

        if parser.options.postlex and parser.options.postlex.always_accept:
            # If postlexer's always_accept is used, we need to recompile the grammar with empty terminals-to-keep
            if not hasattr(parser, 'grammar'):
                raise ConfigurationError('Source grammar not available from cached parser, use cache_grammar=True'
                                         if parser.options.cache else "Source grammar not available!")
            self.tokens, rules, _extra = parser.grammar.compile(parser.options.start, set())
        else:
            self.tokens = list(parser.terminals)
            rules = list(parser.rules)

        self.rules_for_root = defaultdict(list)

        self.rules = list(self._build_recons_rules(rules))
        self.rules.reverse()

        # Choose the best rule from each group of {rule => [rule.alias]}, since we only really need one derivation.
        self.rules = _best_rules_from_group(self.rules)

        self.parser = parser
        self._parser_cache: Dict[str, earley.Parser] = {}

    def _build_recons_rules(self, rules: List[Rule]):
        "Convert tree-parsing/construction rules to tree-matching rules"
        expand1s = {r.origin for r in rules if r.options.expand1}

        aliases = defaultdict(list)
        for r in rules:
            if r.alias:
                aliases[r.origin].append(r.alias)

        rule_names = {r.origin for r in rules}
        nonterminals = {sym for sym in rule_names
                        if sym.name.startswith('_') or sym in expand1s or sym in aliases}

        seen = set()
        for r in rules:
            recons_exp = [sym if sym in nonterminals else Terminal(sym.name)
                          for sym in r.expansion if not is_discarded_terminal(sym)]

            # Skip self-recursive constructs
            if recons_exp == [r.origin] and r.alias is None:
                continue

            sym = NonTerminal(r.alias) if r.alias else r.origin
            rule = make_recons_rule(sym, recons_exp, r.expansion)

            if sym in expand1s and len(recons_exp) != 1:
                self.rules_for_root[sym.name].append(rule)

                if sym.name not in seen:
                    yield make_recons_rule_to_term(sym, sym)
                    seen.add(sym.name)
            else:
                if sym.name.startswith('_') or sym in expand1s:
                    yield rule
                else:
                    self.rules_for_root[sym.name].append(rule)

        for origin, rule_aliases in aliases.items():
            for alias in rule_aliases:
                yield make_recons_rule_to_term(origin, NonTerminal(alias))
            yield make_recons_rule_to_term(origin, origin)

    def match_tree(self, tree: Tree, rulename: str) -> Tree:
        """Match the elements of `tree` to the symbols of rule `rulename`.

        Parameters:
            tree (Tree): the tree node to match
            rulename (str): The expected full rule name (including template args)

        Returns:
            Tree: an unreduced tree that matches `rulename`

        Raises:
            UnexpectedToken: If no match was found.

        Note:
            It's the callers' responsibility to match the tree recursively.
        """
        if rulename:
            # validate
            name, _args = parse_rulename(rulename)
            assert tree.data == name
        else:
            rulename = tree.data

        # TODO: ambiguity?
        try:
            parser = self._parser_cache[rulename]
        except KeyError:
            rules = self.rules + _best_rules_from_group(self.rules_for_root[rulename])

            # TODO pass callbacks through dict, instead of alias?
            callbacks = {rule: rule.alias for rule in rules}
            conf = ParserConf(rules, callbacks, [rulename]) # type: ignore[arg-type]
            parser = earley.Parser(self.parser.lexer_conf, conf, _match, resolve_ambiguity=True)
            self._parser_cache[rulename] = parser

        # find a full derivation
        unreduced_tree: Tree = parser.parse(ChildrenLexer(tree.children), rulename)
        assert unreduced_tree.data == rulename
        return unreduced_tree



from __future__ import annotations

from pathlib import Path

import pytest

from packaging.utils import canonicalize_name

from poetry.core.packages.dependency import Dependency
from poetry.core.packages.dependency_group import DependencyGroup
from poetry.core.packages.directory_dependency import DirectoryDependency
from poetry.core.packages.vcs_dependency import VCSDependency


def create_dependency(
    name: str,
    constraint: str = "*",
    *,
    optional: bool = False,
    extras: tuple[str, ...] = (),
    in_extras: tuple[str, ...] = (),
    allows_prereleases: bool | None = None,
    develop: bool = False,
    source_name: str | None = None,
    marker: str | None = None,
) -> Dependency:
    dep = Dependency(
        name=name,
        constraint=constraint,
        optional=optional,
        extras=extras,
        allows_prereleases=allows_prereleases,
    )
    if in_extras:
        dep._optional = True
        dep._in_extras = [canonicalize_name(extra) for extra in in_extras]
    if develop:
        dep._develop = develop
    if source_name:
        dep.source_name = source_name
    if marker:
        dep.marker = marker
    return dep


@pytest.mark.parametrize("mixed_dynamic", [False, True])
@pytest.mark.parametrize(
    (
        "dependencies",
        "poetry_dependencies",
        "expected_dependencies",
    ),
    [
        ({Dependency("foo", "*", optional=True)}, set(), {"foo"}),
        (set(), {Dependency("bar", "*")}, {"bar"}),
        (set(), {Dependency("bar", "*", optional=True)}, {"bar"}),
        ({Dependency("foo", "*")}, {Dependency("bar", "*")}, {"foo"}),
        (
            {Dependency("foo", "*", optional=True)},
            {Dependency("bar", "*")},
            ({"foo"}, {"foo", "bar"}),
        ),
        (
            {Dependency("foo", "*")},
            {Dependency("bar", "*", optional=True)},
            ({"foo"}, {"foo", "bar"}),
        ),
        (
            {
                Dependency("foo", "*", optional=True),
                Dependency("baz", "*", optional=True),
            },
            {Dependency("bar", "*")},
            ({"foo", "baz"}, {"foo", "bar", "baz"}),
        ),
        (
            {
                Dependency("foo", "*", optional=True),
                Dependency("baz", "*", optional=False),
            },
            {Dependency("bar", "*")},
            {"foo", "baz"},
        ),
        (
            {Dependency("foo", "*", optional=True)},
            {Dependency("bar", "*"), Dependency("baz", "*", optional=True)},
            ({"foo"}, {"foo", "bar"}),
        ),
    ],
)
def test_dependencies(
    dependencies: set[Dependency],
    poetry_dependencies: set[Dependency],
    mixed_dynamic: bool,
    expected_dependencies: set[str] | tuple[set[str], set[str]],
) -> None:
    group = DependencyGroup(name="group", mixed_dynamic=mixed_dynamic)
    group._dependencies = list(dependencies)
    group._poetry_dependencies = list(poetry_dependencies)

    if isinstance(expected_dependencies, tuple):
        expected_dependencies = (
            expected_dependencies[1] if mixed_dynamic else expected_dependencies[0]
        )
    assert {d.name for d in group.dependencies} == expected_dependencies


@pytest.mark.parametrize(
    (
        "initial_dependencies",
        "initial_poetry_dependencies",
        "expected_dependencies",
        "expected_poetry_dependencies",
    ),
    [
        (set(), set(), {"new"}, set()),
        ({"foo"}, set(), {"foo", "new"}, set()),
        (set(), {"bar"}, set(), {"bar", "new"}),
        ({"foo"}, {"bar"}, {"foo", "new"}, {"bar"}),
    ],
)
def test_add_dependency_adds_to_correct_list(
    initial_dependencies: set[str],
    initial_poetry_dependencies: set[str],
    expected_dependencies: set[str],
    expected_poetry_dependencies: set[str],
) -> None:
    group = DependencyGroup(name="group")
    group._dependencies = [
        Dependency(name=name, constraint="*") for name in initial_dependencies
    ]
    group._poetry_dependencies = [
        Dependency(name=name, constraint="*") for name in initial_poetry_dependencies
    ]

    group.add_dependency(Dependency(name="new", constraint="*"))

    assert {d.name for d in group._dependencies} == expected_dependencies
    assert {d.name for d in group._poetry_dependencies} == expected_poetry_dependencies


def test_remove_dependency_removes_from_both_lists() -> None:
    group = DependencyGroup(name="group")
    group.add_dependency(Dependency(name="foo", constraint="*"))
    group.add_dependency(Dependency(name="bar", constraint="*"))
    group.add_dependency(Dependency(name="foo", constraint="*"))
    group.add_poetry_dependency(Dependency(name="baz", constraint="*"))
    group.add_poetry_dependency(Dependency(name="foo", constraint="*"))

    group.remove_dependency("foo")

    assert {d.name for d in group._dependencies} == {"bar"}
    assert {d.name for d in group._poetry_dependencies} == {"baz"}


@pytest.mark.parametrize("mixed_dynamic", [False, True])
@pytest.mark.parametrize(
    (
        "dependencies",
        "poetry_dependencies",
        "expected_dependencies",
    ),
    [
        ([Dependency.create_from_pep_508("foo")], [], [create_dependency("foo")]),
        ([], [Dependency.create_from_pep_508("bar")], [create_dependency("bar")]),
        (
            [create_dependency("foo")],
            [create_dependency("bar")],
            [create_dependency("foo")],
        ),
        (
            [create_dependency("foo", in_extras=("extra1",))],
            [create_dependency("bar")],
            (
                [create_dependency("foo", in_extras=("extra1",))],
                [
                    create_dependency("foo", in_extras=("extra1",)),
                    create_dependency("bar"),
                ],
            ),
        ),
        (
            [create_dependency("foo")],
            [create_dependency("bar", in_extras=("extra1",))],
            (
                [create_dependency("foo")],
                [
                    create_dependency("foo"),
                    create_dependency("bar", in_extras=("extra1",)),
                ],
            ),
        ),
        # refine constraint
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", "<2")],
            [create_dependency("foo", ">=1,<2")],
        ),
        # refine constraint + other dependency
        (
            [
                Dependency.create_from_pep_508("foo>=1"),
                Dependency.create_from_pep_508("bar>=2"),
            ],
            [create_dependency("foo", "<2")],
            [create_dependency("foo", ">=1,<2"), create_dependency("bar", ">=2")],
        ),
        # refine constraint depending on marker
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", "<2", marker="sys_platform == 'win32'")],
            [create_dependency("foo", ">=1,<2", marker="sys_platform == 'win32'")],
        ),
        # allow pre-releases
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", allows_prereleases=True)],
            [create_dependency("foo", ">=1", allows_prereleases=True)],
        ),
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", allows_prereleases=False)],
            [create_dependency("foo", ">=1", allows_prereleases=False)],
        ),
        # directory dependency - develop
        (
            [DirectoryDependency("foo", Path("path/to/foo"))],
            [create_dependency("foo", develop=True)],
            [DirectoryDependency("foo", Path("path/to/foo"), develop=True)],
        ),
        # directory dependency - develop (full spec)
        (
            [DirectoryDependency("foo", Path("path/to/foo"))],
            [DirectoryDependency("foo", Path("path/to/foo"), develop=True)],
            [DirectoryDependency("foo", Path("path/to/foo"), develop=True)],
        ),
        # vcs dependency - develop
        (
            [VCSDependency("foo", "git", "https://example.org/foo")],
            [create_dependency("foo", develop=True)],
            [VCSDependency("foo", "git", "https://example.org/foo", develop=True)],
        ),
        # vcs dependency - develop (full spec)
        (
            [VCSDependency("foo", "git", "https://example.org/foo")],
            [VCSDependency("foo", "git", "https://example.org/foo", develop=True)],
            [VCSDependency("foo", "git", "https://example.org/foo", develop=True)],
        ),
        # replace with directory dependency
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [DirectoryDependency("foo", Path("path/to/foo"), develop=True)],
            [DirectoryDependency("foo", Path("path/to/foo"), develop=True)],
        ),
        # source
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", source_name="src")],
            [create_dependency("foo", ">=1", source_name="src")],
        ),
        # different sources depending on marker
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [
                create_dependency(
                    "foo", source_name="src1", marker="sys_platform == 'win32'"
                ),
                create_dependency(
                    "foo", source_name="src2", marker="sys_platform == 'linux'"
                ),
            ],
            [
                create_dependency(
                    "foo", ">=1", source_name="src1", marker="sys_platform == 'win32'"
                ),
                create_dependency(
                    "foo", ">=1", source_name="src2", marker="sys_platform == 'linux'"
                ),
            ],
        ),
        # pairwise different sources depending on marker
        (
            [
                Dependency.create_from_pep_508("foo>=1; sys_platform == 'win32'"),
                Dependency.create_from_pep_508("foo>=1.1; sys_platform == 'linux'"),
            ],
            [
                create_dependency(
                    "foo", source_name="src1", marker="sys_platform == 'win32'"
                ),
                create_dependency(
                    "foo", source_name="src2", marker="sys_platform == 'linux'"
                ),
            ],
            [
                create_dependency(
                    "foo", ">=1", source_name="src1", marker="sys_platform == 'win32'"
                ),
                create_dependency(
                    "foo", ">=1.1", source_name="src2", marker="sys_platform == 'linux'"
                ),
            ],
        ),
        # enrich only one with source
        (
            [
                Dependency.create_from_pep_508("foo>=1; sys_platform == 'win32'"),
                Dependency.create_from_pep_508("foo>=1.1; sys_platform == 'linux'"),
            ],
            [
                create_dependency(
                    "foo", source_name="src1", marker="sys_platform == 'win32'"
                ),
            ],
            [
                create_dependency(
                    "foo", ">=1", source_name="src1", marker="sys_platform == 'win32'"
                ),
                create_dependency("foo", ">=1.1", marker="sys_platform == 'linux'"),
            ],
        ),
        # extras
        (
            [Dependency.create_from_pep_508("foo[extra1,extra2]")],
            [create_dependency("foo", source_name="src")],
            [create_dependency("foo", source_name="src", extras=("extra1", "extra2"))],
        ),
        (
            [Dependency.create_from_pep_508("foo;extra=='extra1'")],
            [create_dependency("foo", source_name="src", optional=True)],
            [create_dependency("foo", source_name="src", marker="extra == 'extra1'")],
        ),
        (
            [Dependency.create_from_pep_508("foo;extra=='extra1'")],
            [create_dependency("foo", source_name="src")],
            (
                [
                    create_dependency(
                        "foo", source_name="src", marker="extra == 'extra1'"
                    )
                ],
                [
                    create_dependency(
                        "foo", source_name="src", marker="extra == 'extra1'"
                    ),
                    create_dependency("foo", source_name="src"),
                ],
            ),
        ),
        # extras - special
        # root extras do not have an extra marker, they just have set _in_extras!
        (
            [
                create_dependency(
                    "foo", source_name="src", optional=True, in_extras=("extra1",)
                )
            ],
            [create_dependency("foo", source_name="src", optional=True)],
            [
                create_dependency(
                    "foo", source_name="src", optional=True, in_extras=("extra1",)
                )
            ],
        ),
        (
            [
                create_dependency(
                    "foo", source_name="src", optional=True, in_extras=("extra1",)
                )
            ],
            [create_dependency("foo", source_name="src")],
            (
                [
                    create_dependency(
                        "foo", source_name="src", optional=True, in_extras=("extra1",)
                    )
                ],
                [
                    create_dependency(
                        "foo", source_name="src", optional=True, in_extras=("extra1",)
                    ),
                    create_dependency("foo", source_name="src"),
                ],
            ),
        ),
        (
            [
                Dependency.create_from_pep_508("foo;extra!='extra1'"),
                create_dependency("foo", in_extras=("extra1",)),
            ],
            [
                create_dependency("foo", marker="extra!='extra1'", source_name="src1"),
                create_dependency("foo", marker="extra=='extra1'", source_name="src2"),
            ],
            [
                create_dependency("foo", source_name="src1", marker="extra!='extra1'"),
                create_dependency("foo", source_name="src2", in_extras=("extra1",)),
            ],
        ),
        (
            [
                Dependency.create_from_pep_508(
                    "foo;extra!='extra1' and extra!='extra2'"
                ),
                create_dependency("foo", in_extras=("extra1", "extra2")),
            ],
            [
                create_dependency(
                    "foo",
                    marker="extra!='extra1' and extra!='extra2'",
                    source_name="src1",
                ),
                create_dependency(
                    "foo",
                    marker="extra=='extra1' or extra=='extra2'",
                    source_name="src2",
                ),
            ],
            [
                create_dependency(
                    "foo",
                    source_name="src1",
                    marker="extra!='extra1' and extra!='extra2'",
                ),
                create_dependency(
                    "foo", source_name="src2", in_extras=("extra1", "extra2")
                ),
            ],
        ),
        (
            [
                create_dependency(
                    "foo", marker="extra!='extra2'", in_extras=("extra1",)
                ),
                create_dependency(
                    "foo", marker="extra!='extra1'", in_extras=("extra2",)
                ),
            ],
            [
                create_dependency(
                    "foo",
                    marker="extra!='extra2' and extra=='extra1'",
                    source_name="src1",
                ),
                create_dependency(
                    "foo",
                    marker="extra!='extra1' and extra=='extra2'",
                    source_name="src2",
                ),
            ],
            [
                create_dependency(
                    "foo",
                    source_name="src1",
                    marker="extra!='extra2'",
                    in_extras=("extra1",),
                ),
                create_dependency(
                    "foo",
                    source_name="src2",
                    marker="extra!='extra1'",
                    in_extras=("extra2",),
                ),
            ],
        ),
    ],
)
def test_dependencies_for_locking(
    dependencies: list[Dependency],
    poetry_dependencies: list[Dependency],
    mixed_dynamic: bool,
    expected_dependencies: list[Dependency] | tuple[list[Dependency], list[Dependency]],
) -> None:
    group = DependencyGroup(name="group", mixed_dynamic=mixed_dynamic)
    group._dependencies = dependencies
    group._poetry_dependencies = poetry_dependencies

    if isinstance(expected_dependencies, tuple):
        expected_dependencies = (
            expected_dependencies[1] if mixed_dynamic else expected_dependencies[0]
        )

    assert group.dependencies_for_locking == expected_dependencies
    # explicitly check attributes that are not considered in __eq__
    assert [d.allows_prereleases() for d in group.dependencies_for_locking] == [
        d.allows_prereleases() for d in expected_dependencies
    ]
    assert [d.source_name for d in group.dependencies_for_locking] == [
        d.source_name for d in expected_dependencies
    ]
    assert [d.marker for d in group.dependencies_for_locking] == [
        d.marker for d in expected_dependencies
    ]
    assert [d._develop for d in group.dependencies_for_locking] == [
        d._develop for d in expected_dependencies
    ]
    assert [d.in_extras for d in group.dependencies_for_locking] == [
        d.in_extras for d in expected_dependencies
    ]
    assert [d.is_optional() for d in group.dependencies_for_locking] == [
        d.is_optional() for d in expected_dependencies
    ]


@pytest.mark.parametrize(
    (
        "dependencies",
        "poetry_dependencies",
    ),
    [
        (
            [Dependency.create_from_pep_508("foo>=1")],
            [create_dependency("foo", "<1")],
        ),
        (
            [DirectoryDependency("foo", Path("path/to/foo"))],
            [VCSDependency("foo", "git", "https://example.org/foo")],
        ),
    ],
)
def test_dependencies_for_locking_failure(
    dependencies: list[Dependency],
    poetry_dependencies: list[Dependency],
) -> None:
    group = DependencyGroup(name="group")
    group._dependencies = dependencies
    group._poetry_dependencies = poetry_dependencies

    with pytest.raises(ValueError):
        _ = group.dependencies_for_locking


@pytest.mark.parametrize(
    ["pretty_name", "canonicalized_name"],
    [
        ("Group", "group"),
        ("foo_bar", "foo-bar"),
        ("Foo-Bar", "foo-bar"),
    ],
)
def test_dependency_group_use_canonicalize_name(
    pretty_name: str, canonicalized_name: str
) -> None:
    group = DependencyGroup(pretty_name)
    assert group.name == canonicalized_name
    assert group.pretty_name == pretty_name


def test_include_dependency_groups() -> None:
    group = DependencyGroup(name="group")
    group.add_poetry_dependency(
        Dependency(name="foo", constraint="*", groups=["group"])
    )

    group_2 = DependencyGroup(name="group2")
    group_2.add_poetry_dependency(
        Dependency(name="bar", constraint="*", groups=["group2"])
    )

    # This will resolve to a more precise constraint in dependencies_for_locking.
    group_3 = DependencyGroup(name="group3")
    group_3._dependencies = [Dependency(name="baz", constraint="<2", groups=["group3"])]
    group_3._poetry_dependencies = [
        Dependency(name="baz", constraint=">=1", groups=["group3"])
    ]

    group.include_dependency_group(group_2)
    group.include_dependency_group(group_3)

    assert [dep.name for dep in group.dependencies] == ["foo", "bar", "baz"]
    assert [dep.name for dep in group.dependencies_for_locking] == ["foo", "bar", "baz"]
    assert [dep.pretty_constraint for dep in group.dependencies] == ["*", "*", "<2"]
    assert [dep.pretty_constraint for dep in group.dependencies_for_locking] == [
        "*",
        "*",
        ">=1,<2",
    ]
    for dep in group.dependencies:
        assert dep.groups == {"group"}

    assert [dep.name for dep in group_2.dependencies] == ["bar"]
    assert [dep.name for dep in group_2.dependencies_for_locking] == ["bar"]
    for dep in group_2.dependencies:
        assert dep.groups == {"group2"}

    assert [dep.name for dep in group_3.dependencies] == ["baz"]
    assert [dep.name for dep in group_3.dependencies_for_locking] == ["baz"]
    assert [dep.pretty_constraint for dep in group_3.dependencies] == ["<2"]
    assert [dep.pretty_constraint for dep in group_3.dependencies_for_locking] == [
        ">=1,<2"
    ]
    for dep in group_3.dependencies:
        assert dep.groups == {"group3"}


def test_include_empty_dependency_group() -> None:
    group_main = DependencyGroup(name="group")
    group_main.add_poetry_dependency(
        Dependency(name="foo", constraint="*", groups=["group"])
    )

    empty_group = DependencyGroup(name="empty")
    # Include empty dependency group
    group_main.include_dependency_group(empty_group)

    # Assert that including an empty group does not affect group_main's dependencies.
    assert [dep.name for dep in group_main.dependencies] == ["foo"]
    assert [dep.name for dep in group_main.dependencies_for_locking] == ["foo"]


@pytest.mark.parametrize("group_name", ["group_2", "Group-2", "group-2"])
def test_include_dependency_group_raise_if_including_itself(group_name: str) -> None:
    group = DependencyGroup(name="group-2")
    group.add_poetry_dependency(
        Dependency(name="foo", constraint="*", groups=["group"])
    )

    with pytest.raises(
        ValueError, match="Cannot include the dependency group to itself"
    ):
        group.include_dependency_group(DependencyGroup(name=group_name))


@pytest.mark.parametrize("group_name", ["group_2", "Group-2", "group-2"])
def test_include_dependency_group_raise_if_already_included(group_name: str) -> None:
    group = DependencyGroup(name="group")
    group.add_poetry_dependency(
        Dependency(name="foo", constraint="*", groups=["group"])
    )

    group_2 = DependencyGroup(name="group_2")
    group_2.add_poetry_dependency(
        Dependency(name="bar", constraint="*", groups=["group2"])
    )

    group.include_dependency_group(group_2)

    group_3 = DependencyGroup(name=group_name)
    group_3.add_poetry_dependency(
        Dependency(name="bar", constraint="*", groups=["group2"])
    )

    with pytest.raises(
        ValueError, match=f"Dependency group {group_name} is already included"
    ):
        group.include_dependency_group(group_3)



from __future__ import annotations

from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from poetry.core.constraints.version.version_constraint import VersionConstraint


if TYPE_CHECKING:
    from poetry.core.constraints.version.version import Version


class VersionRangeConstraint(VersionConstraint):
    @property
    @abstractmethod
    def min(self) -> Version | None:
        raise NotImplementedError

    @property
    @abstractmethod
    def max(self) -> Version | None:
        raise NotImplementedError

    @property
    @abstractmethod
    def include_min(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def include_max(self) -> bool:
        raise NotImplementedError

    @property
    def allowed_min(self) -> Version | None:
        # That is a bit inaccurate because
        # 1) The exclusive ordered comparison >V MUST NOT allow a post-release
        #    of the given version unless V itself is a post release.
        # 2) The exclusive ordered comparison >V MUST NOT match
        #    a local version of the specified version.
        # https://peps.python.org/pep-0440/#exclusive-ordered-comparison
        # However, there is no specific min greater than the greatest post release
        # or greatest local version identifier. These cases have to be handled by
        # the callers of allowed_min.
        return self.min

    @cached_property
    def allowed_max(self) -> Version | None:
        if self.max is None:
            return None

        if self.include_max or self.max.is_unstable():
            return self.max

        if self.min == self.max and (self.include_min or self.include_max):
            # this is an equality range
            return self.max

        # The exclusive ordered comparison <V MUST NOT allow a pre-release
        # of the specified version unless the specified version is itself a pre-release.
        # https://peps.python.org/pep-0440/#exclusive-ordered-comparison
        return self.max.first_devrelease()

    def has_upper_bound(self) -> bool:
        return self.max is not None

    def allows_lower(self, other: VersionRangeConstraint) -> bool:
        _this, _other = self.allowed_min, other.allowed_min

        if _this is None:
            return _other is not None

        if _other is None:
            return False

        if _this < _other:
            return True

        if _this > _other:
            return False

        return self.include_min and not other.include_min

    def allows_higher(self, other: VersionRangeConstraint) -> bool:
        _this, _other = self.allowed_max, other.allowed_max

        if _this is None:
            return _other is not None

        if _other is None:
            return False

        if _this < _other:
            return False

        if _this > _other:
            return True

        return self.include_max and not other.include_max

    def is_strictly_lower(self, other: VersionRangeConstraint) -> bool:
        _this, _other = self.allowed_max, other.allowed_min

        if _this is None or _other is None:
            return False

        if _this < _other:
            return True

        if _this > _other:
            return False

        return not (self.include_max and other.include_min)

    def is_strictly_higher(self, other: VersionRangeConstraint) -> bool:
        return other.is_strictly_lower(self)

    def is_adjacent_to(self, other: VersionRangeConstraint) -> bool:
        if self.max != other.min:
            return False

        return (self.include_max and not other.include_min) or (
            not self.include_max and other.include_min
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from pathlib import Path

    from lark import Lark
    from lark import Tree


class Parser:
    def __init__(
        self, grammar: Path, parser: str = "lalr", debug: bool = False
    ) -> None:
        self._grammar = grammar
        self._parser = parser
        self._debug = debug
        self._lark: Lark | None = None

    def parse(self, text: str, **kwargs: Any) -> Tree:
        from lark import Lark

        if self._lark is None:
            self._lark = Lark.open(
                grammar_filename=self._grammar, parser=self._parser, debug=self._debug
            )

        return self._lark.parse(text=text, **kwargs)
